import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import timedelta
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    runtime_checkable,
)

from golem.exceptions import GolemException
from golem.resources import (
    Activity,
    Agreement,
    Allocation,
    DemandData,
    Proposal,
    ProposalData,
    Script,
)
from golem.resources.activity import commands
from golem.utils.typing import MaybeAwaitable

logger = logging.getLogger(__name__)


class Batch:
    def __init__(
        self, activity: Activity, timeout: Optional[Union[timedelta, float]] = None
    ) -> None:
        self._script = Script()
        self._activity = activity
        self._call_timeout = timeout

    def deploy(self, deploy_args: Optional[commands.ArgsDict] = None):
        self._script.add_command(commands.Deploy(deploy_args))

    def start(self):
        self._script.add_command(commands.Start())

    def send_file(self, src_path: str, dst_path: str):
        self._script.add_command(commands.SendFile(src_path, dst_path))

    def download_file(self, src_path: str, dst_path: str):
        self._script.add_command(commands.DownloadFile(src_path, dst_path))

    def run(
        self,
        command: Union[str, List[str]],
        *,
        shell: Optional[bool] = None,
        shell_cmd: str = "/bin/sh",
    ):
        self._script.add_command(commands.Run(command, shell=shell, shell_cmd=shell_cmd))

    async def __call__(self):
        pooling_batch = await self._activity.execute_script(self._script)
        return await pooling_batch.wait(timeout=self._call_timeout)


class WorkContext:
    def __init__(self, activity: Activity):
        self._activity = activity

    async def deploy(
        self, deploy_args: Optional[commands.ArgsDict] = None, timeout: Optional[timedelta] = None
    ):
        pooling_batch = await self._activity.execute_commands(commands.Deploy(deploy_args))
        await pooling_batch.wait(timeout)

    async def start(self):
        pooling_batch = await self._activity.execute_commands(commands.Start())
        await pooling_batch.wait()

    async def terminate(self):
        await self._activity.destroy()

    async def run(
        self,
        command: Union[str, List[str]],
        *,
        shell: Optional[bool] = None,
        shell_cmd: str = "/bin/sh",
        timeout: Optional[float] = None,
    ):
        return await self._activity.execute_commands(
            commands.Run(command, shell=shell, shell_cmd=shell_cmd), timeout=timeout
        )

    async def create_batch(self) -> Batch:
        return Batch(self._activity)

    @property
    def activity(self) -> Activity:
        return self._activity

    async def get_provider_id(self):
        """Get the node id of the provider running this context."""
        return await self.activity.agreement.proposal.get_provider_id()

    async def get_provider_name(self):
        """Get the node name of the provider running this context."""
        return await self.activity.agreement.proposal.get_provider_name()


@dataclass
class WorkResult:
    result: Optional[Any] = None
    exception: Optional[Exception] = None
    extras: Dict = field(default_factory=dict)


WORK_PLUGIN_FIELD_NAME = "_work_plugins"

Work = Callable[[WorkContext], Awaitable[Optional[WorkResult]]]

DoWorkCallable = Callable[[Work], Awaitable[WorkResult]]


class ManagerEvent(ABC):
    pass


class ManagerException(GolemException):
    pass


class ManagerPluginException(ManagerException):
    pass


class Manager(ABC):
    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.stop()

    async def start(self) -> None:
        pass

    async def stop(self) -> None:
        pass


TPlugin = TypeVar("TPlugin")


class NetworkManager(Manager, ABC):
    pass


class PaymentManager(Manager, ABC):
    @abstractmethod
    async def get_allocation(self) -> Allocation:
        ...


class DemandManager(Manager, ABC):
    @abstractmethod
    async def get_initial_proposal(self) -> Proposal:
        ...


class ProposalManager(Manager):
    @abstractmethod
    async def get_draft_proposal(self) -> Proposal:
        ...


class AgreementManager(Manager, ABC):
    @abstractmethod
    async def get_agreement(self) -> Agreement:
        ...


class ActivityManager(Manager, ABC):
    @abstractmethod
    async def get_activity(self) -> Activity:
        ...


class WorkManager(Manager, ABC):
    pass


class RejectProposal(ManagerPluginException):
    pass


@runtime_checkable
class ProposalNegotiator(Protocol):
    def __call__(
        self, demand_data: DemandData, proposal_data: ProposalData
    ) -> MaybeAwaitable[Optional[RejectProposal]]:
        ...


class ProposalManagerPlugin(ABC):
    _get_proposal: Callable[[], Awaitable[Proposal]]

    def set_proposal_callback(self, get_proposal: Callable[[], Awaitable[Proposal]]):
        self._get_proposal = get_proposal

    @abstractmethod
    async def get_proposal(self) -> Proposal:
        ...

    async def start(self) -> None:
        pass

    async def stop(self) -> None:
        pass


ProposalScoringResult = Sequence[Optional[float]]


@runtime_checkable
class ProposalScorer(Protocol):
    def __call__(
        self, proposals_data: Sequence[ProposalData]
    ) -> MaybeAwaitable[ProposalScoringResult]:
        ...


ScorerWithOptionalWeight = Union[ProposalScorer, Tuple[float, ProposalScorer]]


@runtime_checkable
class WorkManagerPlugin(Protocol):
    def __call__(self, do_work: DoWorkCallable) -> DoWorkCallable:
        ...


# TODO: Make consistent naming on functions in arguments in whole project - callable or func
PricingCallable = Callable[[ProposalData], Optional[float]]



from golem.managers.proposal.plugins.scoring.map import MapScore
from golem.managers.proposal.plugins.scoring.mixins import ProposalScoringMixin
from golem.managers.proposal.plugins.scoring.pricings import (
    LinearAverageCostPricing,
    LinearPerCpuAverageCostPricing,
)
from golem.managers.proposal.plugins.scoring.property_value_lerp import PropertyValueLerpScore
from golem.managers.proposal.plugins.scoring.random import RandomScore
from golem.managers.proposal.plugins.scoring.scoring_buffer import ProposalScoringBuffer

__all__ = (
    "MapScore",
    "ProposalScoringMixin",
    "LinearAverageCostPricing",
    "LinearPerCpuAverageCostPricing",
    "PropertyValueLerpScore",
    "RandomScore",
    "ProposalScoringBuffer",
)



import logging
from typing import Awaitable, Callable

from golem.managers.base import (
    WORK_PLUGIN_FIELD_NAME,
    DoWorkCallable,
    Work,
    WorkContext,
    WorkManagerPlugin,
    WorkResult,
)
from golem.managers.mixins import PluginsMixin
from golem.resources import Activity
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)


class WorkManagerPluginsMixin(PluginsMixin[WorkManagerPlugin]):
    @trace_span()
    def _apply_plugins_from_manager(self, do_work: DoWorkCallable) -> DoWorkCallable:
        do_work_with_plugins = do_work

        for plugin in self._plugins:
            do_work_with_plugins = plugin(do_work_with_plugins)

        return do_work_with_plugins

    @trace_span()
    def _apply_plugins_from_work(self, do_work: DoWorkCallable, work: Work) -> DoWorkCallable:
        work_plugins = getattr(work, WORK_PLUGIN_FIELD_NAME, [])

        if not work_plugins:
            return do_work

        do_work_with_plugins = do_work

        for plugin in work_plugins:
            do_work_with_plugins = plugin(do_work_with_plugins)

        return do_work_with_plugins

    @trace_span()
    async def _do_work_with_plugins(self, do_work: DoWorkCallable, work: Work) -> WorkResult:
        do_work_with_plugins = self._apply_plugins_from_manager(do_work)
        do_work_with_plugins = self._apply_plugins_from_work(do_work_with_plugins, work)
        return await do_work_with_plugins(work)


class WorkManagerDoWorkMixin:
    _get_activity: Callable[[], Awaitable[Activity]]

    async def _do_work(self, work: Work) -> WorkResult:
        activity = await self._get_activity()
        work_context = WorkContext(activity)
        try:
            work_result = await work(work_context)
        except Exception as e:
            work_result = WorkResult(exception=e)
        else:
            if not isinstance(work_result, WorkResult):
                work_result = WorkResult(result=work_result)
        await activity.destroy()
        return work_result



import asyncio
import collections
from typing import Callable, Deque


class Waiter:
    """Class similar to `asyncio.Event` but valueless and with notify interface similar to \
    `asyncio.Condition`.

    Note: Developed to support `golem.utils.asyncio.buffer`, but finally not used.
    """

    def __init__(self) -> None:
        self._waiters: Deque[asyncio.Future] = collections.deque()
        self._loop = asyncio.get_event_loop()

    async def wait_for(self, predicate: Callable[[], bool]) -> None:
        """Check if predicate is true and return immediately, or await until it becomes true."""

        result = predicate()

        while not result:
            await self._wait()
            result = predicate()

            if not result:
                # as last `._wait()` call woken us up but predicate is still false, lets give a
                # chance another `.wait_for()` pending call.
                self._notify_first()

    def _notify_first(self) -> None:
        try:
            first_waiter = self._waiters[0]
        except IndexError:
            return

        if not first_waiter.done():
            first_waiter.set_result(None)

    async def _wait(self) -> None:
        future = self._loop.create_future()
        self._waiters.append(future)
        try:
            await future
        finally:
            self._waiters.remove(future)

    def notify(self, count=1) -> None:
        """Notify given amount of `.wait_for()` calls to check its predicates."""

        notified = 0
        for waiter in self._waiters:
            if count <= notified:
                break

            if not waiter.done():
                waiter.set_result(None)
                notified += 1



import asyncio
import shlex
from abc import ABC, abstractmethod
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

from ya_activity import models

from golem.utils.storage import Destination, Source
from golem.utils.storage.gftp import GftpProvider

ArgsDict = Mapping[str, Union[str, List, Dict[str, Any]]]


class Command(ABC):
    def text(self) -> Dict[str, ArgsDict]:
        return {self.command_name: self.args_dict()}

    @property
    def command_name(self) -> str:
        return type(self).__name__.lower()

    @abstractmethod
    def args_dict(self) -> ArgsDict:
        raise NotImplementedError

    async def before(self) -> None:
        pass

    async def after(self) -> None:
        pass


class Script:
    """A helper class for executing multiple commands in a single batch.

    Details: :any:`execute_script`.
    """

    def __init__(self) -> None:
        self._commands: List[Command] = []
        self._futures: List[asyncio.Future[models.ExeScriptCommandResult]] = []

    @property
    def commands(self) -> List[Command]:
        return self._commands.copy()

    @property
    def futures(self) -> "List[asyncio.Future[models.ExeScriptCommandResult]]":
        return self._futures.copy()

    def add_command(self, command: Command) -> "asyncio.Future[models.ExeScriptCommandResult]":
        """Add a :any:`Command` to the script.

        Returns an awaitable that will (after a call to :any:`execute_script`):

        * Return the result of the command once it finished succesfully
        * Raise :any:`CommandFailed` if the command failed
        * Raise :any:`CommandCancelled` if any previous command in the same batch failed

        """
        self._commands.append(command)
        fut: asyncio.Future[models.ExeScriptCommandResult] = asyncio.Future()
        self._futures.append(fut)
        return fut


class Deploy(Command):
    """Executes `deploy()` in the exeunit."""

    def __init__(self, args_dict: Optional[ArgsDict] = None):
        self._args_dict = args_dict or {}

    def args_dict(self) -> ArgsDict:
        return self._args_dict


class Start(Command):
    """Executes `start()` in the exeunit."""

    def args_dict(self) -> ArgsDict:
        return {}


class Run(Command):
    """Executes `run()` in the exeunit."""

    def __init__(
        self,
        command: Union[str, List[str]],
        *,
        shell: Optional[bool] = None,
        shell_cmd: str = "/bin/sh",
    ):
        """Init Run.

        :param command: Either a list `[entry_point, *args]` or a string.
        :param shell: If True, command will be passed as a string to "/bin/sh -c".
            Default value is True if `command` is a string and False if it is a list.
        :param shell_cmd: Shell command, matters only in `shell` is True.

        Examples::

            Run(["/bin/echo", "foo"])                       # /bin/echo "foo"
            Run("echo foo")                                 # /bin/sh -c "echo foo"
            Run(["echo", "foo"], shell=True)                # /bin/sh -c "echo foo"
            Run(["/bin/echo", "foo", ">", "/my_volume/x"])  # /bin/echo "foo" ">" "/my_volume/x"
                                                            # (NOTE: this is usually **not** the
                                                            # intended effect)
            Run("echo foo > /my_volume/x")                  # /bin/sh -c "echo foo > /my_volume/x"
                                                            # (This is better)
        """
        self.entry_point, self.args = self._resolve_init_args(command, shell, shell_cmd)

    def args_dict(self) -> ArgsDict:
        return {
            "entry_point": self.entry_point,
            "args": self.args,
            "capture": {
                "stdout": {
                    "stream": {},
                },
                "stderr": {
                    "stream": {},
                },
            },
        }

    @staticmethod
    def _resolve_init_args(
        command: Union[str, List[str]], shell: Optional[bool], shell_cmd: str
    ) -> Tuple[str, List[str]]:
        if shell is None:
            shell = isinstance(command, str)

        if shell:
            command_str = command if isinstance(command, str) else shlex.join(command)
            entry_point = shell_cmd
            args = ["-c", command_str]
        else:
            command_list = command if isinstance(command, list) else shlex.split(command)
            entry_point, *args = command_list

        if len(entry_point.split()) > 1:
            raise ValueError(f"Whitespaces in entry point '{entry_point}' are forbidden")

        return entry_point, args


class SendFile(Command):
    """Sends a local file to the exeunit."""

    command_name = "transfer"

    def __init__(self, src_path: str, dst_path: str):
        """Init SendFile.

        :param src_path: Name of the local file.
        :param dst_path: Remote (provider-side) path where the file will be saved.
            Usually this path should be under a directory specified as a VOLUME in the image.
        """
        self.src_path = src_path
        self.dst_path = dst_path

        self._tmp_dir = TemporaryDirectory()
        self._gftp = GftpProvider(tmpdir=self._tmp_dir.name)
        self._source: Optional[Source] = None

    async def before(self) -> None:
        self._source = await self._gftp.upload_file(Path(self.src_path))

    async def after(self) -> None:
        assert self._source is not None
        await self._gftp.release_source(self._source)

    def args_dict(self) -> ArgsDict:
        assert self._source is not None
        return {
            "from": self._source.download_url,
            "to": f"container:{self.dst_path}",
        }


class DownloadFile(Command):
    """Downloads a file from the exeunit."""

    command_name = "transfer"

    def __init__(self, src_path: str, dst_path: str):
        """Init DownloadFile.

        :param src_path: Remote (provider-side) name of the file.
        :param dst_path: Path in the local filesystem where the file will be saved.
        """
        self.src_path = src_path
        self.dst_path = dst_path

        self._tmp_dir = TemporaryDirectory()
        self._gftp = GftpProvider(tmpdir=self._tmp_dir.name)
        self._destination: Optional[Destination] = None

    async def before(self) -> None:
        self._destination = await self._gftp.new_destination(Path(self.dst_path))

    async def after(self) -> None:
        assert self._destination is not None
        await self._destination.download_file(Path(self.dst_path))

    def args_dict(self) -> ArgsDict:
        assert self._destination is not None
        return {
            "from": f"container:{self.src_path}",
            "to": self._destination.upload_url,
        }



import asyncio
import logging
from typing import Awaitable, Callable, List

from golem.managers.base import AgreementManager
from golem.node import GolemNode
from golem.resources import ActivityClosed, Agreement, Proposal
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)


class DefaultAgreementManager(AgreementManager):
    def __init__(
        self,
        golem: GolemNode,
        get_draft_proposal: Callable[[], Awaitable[Proposal]],
        *args,
        **kwargs,
    ):
        self._get_draft_proposal = get_draft_proposal
        self._event_bus = golem.event_bus

        self._agreements: List[Agreement] = []

        super().__init__(*args, **kwargs)

    @trace_span("Stopping DefaultAgreementManager", log_level=logging.INFO)
    async def stop(self) -> None:
        if self._agreements:
            await asyncio.gather(*[agreement.close_all() for agreement in self._agreements])
        else:
            logger.info("All agreements are already terminated")

    @trace_span("Getting agreement", show_results=True, log_level=logging.INFO)
    async def get_agreement(self) -> Agreement:
        while True:
            proposal = await self._get_draft_proposal()
            try:
                agreement = await proposal.create_agreement()
                await agreement.confirm()
                await agreement.wait_for_approval()
            except Exception as e:
                logger.debug(f"Creating agreement failed with `{e}`. Retrying...")
            else:
                # TODO: Support removing callback on resource close
                await self._event_bus.on_once(
                    ActivityClosed,
                    self._terminate_agreement,
                    lambda event: event.resource.parent.id == agreement.id,
                )
                self._agreements.append(agreement)
                return agreement

    @trace_span(show_arguments=True)
    async def _terminate_agreement(self, event: ActivityClosed) -> None:
        # TODO ensure agreement it is terminated on SIGINT
        agreement: Agreement = event.resource.parent
        try:
            self._agreements.remove(agreement)
        except ValueError:
            logger.warning(f"Agreement `{agreement}` not found when removing it from tracked list")
        await agreement.close_all()

        logger.info(f"Agreement `{agreement}` closed")



import asyncio
from ipaddress import IPv4Address, IPv4Network, IPv6Address, IPv6Network, ip_network
from typing import TYPE_CHECKING, Dict, List, Optional, TypedDict, Union

from ya_net import RequestorApi, models

from golem.resources.base import _NULL, Resource, api_call_wrapper
from golem.resources.network.events import NetworkClosed, NewNetwork
from golem.resources.network.exceptions import NetworkFull

if TYPE_CHECKING:
    from golem.node import GolemNode

IpAddress = Union[IPv4Address, IPv6Address]
IpNetwork = Union[IPv4Network, IPv6Network]


class DeployArgsType(TypedDict):
    id: str
    ip: str
    mask: Optional[str]
    nodeIp: str
    nodes: Dict[str, str]


class Network(Resource[RequestorApi, models.Network, _NULL, _NULL, _NULL]):
    """A single vpn on the Golem Network.

    Sample usage::

        activity: Activity
        network = await golem.create_network()

        provider_id = activity.parent.parent.data.issuer_id
        ip = await network.create_node(provider_id)

        deploy_args = {"net": [network.deploy_args(ip)]}
        await activity.execute_commands(commands.Deploy(deploy_args))

        #   activity can now be accessed from other nodes in the network

    """

    def __init__(self, golem_node: "GolemNode", id_: str, data: models.Network):
        super().__init__(golem_node, id_, data)
        asyncio.create_task(golem_node.event_bus.emit(NewNetwork(self)))

        self._create_node_lock = asyncio.Lock()
        self._ip_network: IpNetwork = ip_network(data.ip, strict=False)
        self._all_ips = [str(ip) for ip in self._ip_network.hosts()]

        self._requestor_ips: List[str] = []
        self._nodes: Dict[str, str] = {}

    @property
    def network_address(self) -> str:
        return str(self._ip_network.network_address)

    @classmethod
    @api_call_wrapper()
    async def create(
        cls, golem_node: "GolemNode", ip: str, mask: Optional[str], gateway: Optional[str]
    ) -> "Network":
        api = cls._get_api(golem_node)
        in_data = models.Network(ip=ip, mask=mask, gateway=gateway)
        created_data = await api.create_network(in_data)
        return cls(golem_node, created_data.id, created_data)

    @api_call_wrapper(ignore_status_codes=(404,))
    async def remove(self) -> None:
        """Remove the network."""
        await self.api.remove_network(self.id)
        await self.node.event_bus.emit(NetworkClosed(self))

    @api_call_wrapper()
    async def create_node(self, provider_id: str, node_ip: Optional[str] = None) -> str:
        """Add a node to the network.

        :param provider_id: ID of the provider who hosts the :any:`Activity` that will
            be connected to the new node.
        :param node_ip: IP of the node in the network, by default next free IP will be used.
        :return: IP of the node in the network.
        """
        #   Q: Why is there no `Node` class?
        #   A: Mostly because yagna nodes don't have proper IDs (they are just provider_ids), and
        #      this is strongly against the current golem object model (e.g. what if we want to
        #      have the same provider in muliple networks? Nodes would share the same id, but are
        #      totally diferent objects). We could bypass this by having some internal ids
        #      (e.g. network_id-provider_id, or just uuid), but this would not be pretty and there's
        #      no gain from having a Node object either way. This might change in the future.
        async with self._create_node_lock:
            if node_ip is None:
                node_ip = self._next_free_ip()
            assert node_ip is not None  # mypy, why?

            data = models.Node(id=provider_id, ip=node_ip)  # type: ignore  # mypy, why?
            await self.api.add_node(self.id, data)

            self._nodes[node_ip] = provider_id
            return node_ip

    @api_call_wrapper()
    async def refresh_nodes(self) -> None:
        """Propagates the information about created nodes to the network.

        TODO: more precise explanation maybe?
        """
        tasks = []
        for ip, provider_id in self._nodes.items():
            data = models.Node(id=provider_id, ip=ip)  # type: ignore  # mypy, why?
            tasks.append(self.api.add_node(self.id, data))
        await asyncio.gather(*tasks)

    def deploy_args(self, ip: str) -> DeployArgsType:
        """Get a data structure that should be passed to :any:`Deploy`.

        This makes an :any:`Activity` a part of the network.

        :param ip: IP of the previously created network node.

        Sample usage::

            deploy_args = {"net": [network.deploy_args(ip)]}
            await activity.execute_commands(commands.Deploy(deploy_args))

        """
        return {
            "id": self.id,
            "ip": self.network_address,
            "mask": self.data.mask,
            "nodeIp": ip,
            "nodes": self._nodes,
        }

    @api_call_wrapper()
    async def add_requestor_ip(self, ip: Optional[str]) -> None:
        async with self._create_node_lock:
            if ip is None:
                ip = self._next_free_ip()
            self._requestor_ips.append(ip)
        await self.api.add_address(self.id, models.Address(ip))

    @property
    def _current_ips(self) -> List[str]:
        return self._requestor_ips + list(self._nodes)

    def _next_free_ip(self) -> str:
        try:
            return next(ip for ip in self._all_ips if ip not in self._current_ips)
        except StopIteration:
            raise NetworkFull(self)

    @classmethod
    def _id_field_name(cls) -> str:
        #   All ya_client models have id fields that include model name, e.g.
        #   Allocation.allocation_id, Demand.demand_id etc, but we have Network.id
        return "id"



from golem.resources.proposal.data import ProposalData, ProposalId
from golem.resources.proposal.events import NewProposal, ProposalClosed, ProposalDataChanged
from golem.resources.proposal.pipeline import default_create_agreement, default_negotiate
from golem.resources.proposal.proposal import Proposal

__all__ = (
    "Proposal",
    "ProposalData",
    "ProposalId",
    "NewProposal",
    "ProposalDataChanged",
    "ProposalClosed",
    "default_negotiate",
    "default_create_agreement",
)



import asyncio
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Set

from golem.resources import NewDebitNote, NewInvoice

if TYPE_CHECKING:
    from golem.node import GolemNode
    from golem.resources import Allocation, NewAgreement


class DefaultPaymentHandler:
    """Accepts all incoming debit_notes and invoices.

    Calls `get_data(force=True)` on invoices/debit notes after they are accepted,
    so appropriate :any:`ResourceDataChanged` event is emitted.

    Usage::

        async with GolemNode() as golem:
            allocation = await golem.create_allocation(BUDGET)
            payment_handler = DefaultPaymentHandler(golem, allocation)

            try:
                #   ... interact with the Golem Network ...
            finally:
                await payment_handler.terminate_agreements()
                await payment_handler.wait_for_invoices()

    """

    def __init__(self, node: "GolemNode", allocation: "Allocation"):
        """Init DefaultPaymentHandler.

        :param node: Debit notes/invoices received by this node will be accepted.
            :any:`DefaultPaymentHandler` will only work if the :any:`GolemNode` was started with
            `collect_payment_events = True`.
        :param allocation: Allocation that will be used to accept debit notes/invoices.
        """

        # FIXME: Resolve local import due to cyclic imports
        from golem.resources import Agreement

        self._node = node
        self.allocation = allocation
        self._agreements: Set[Agreement] = set()

    async def start(self):
        # FIXME: Add event_bus.off

        await self._node.event_bus.on(NewAgreement, self.on_agreement)
        await self._node.event_bus.on(NewInvoice, self.on_invoice)
        await self._node.event_bus.on(NewDebitNote, self.on_debit_note)

    async def on_agreement(self, event: "NewAgreement") -> None:
        # FIXME: Resolve local import due to cyclic imports

        self._agreements.add(event.resource)

    async def on_invoice(self, event: NewInvoice) -> None:
        invoice = event.resource

        if (await invoice.get_data(force=True)).status == "RECEIVED":
            await invoice.accept_full(self.allocation)
            await invoice.get_data(force=True)

    async def on_debit_note(self, event: NewDebitNote) -> None:
        debit_note = event.resource

        if (await debit_note.get_data(force=True)).status == "RECEIVED":
            await debit_note.accept_full(self.allocation)
            await debit_note.get_data(force=True)

    async def terminate_agreements(self) -> None:
        """Terminate all agreements and activities.

        This is intended to be used just before :any:`wait_for_invoices`.
        """
        await asyncio.gather(*[agreement.close_all() for agreement in self._agreements])

    async def wait_for_invoices(self, timeout: float = 5) -> None:
        """Wait for invoices for all agreements and accept them.

        :param timeout: Maximum wait time in seconds.
        """
        stop = datetime.now() + timedelta(seconds=timeout)
        while datetime.now() < stop and any(
            agreement.invoice is None or agreement.invoice.data.status == "RECEIVED"
            for agreement in self._agreements
        ):
            await asyncio.sleep(0.1)

        missing_invoices = [
            agreement for agreement in self._agreements if agreement.invoice is None
        ]
        if missing_invoices:
            missing_invoices_str = ", ".join([str(agreement) for agreement in missing_invoices])
            print(
                f"wait_for_invoices timed out, agreements without invoices: {missing_invoices_str}"
            )



from typing import Any, Awaitable, Callable, Optional, Type, TypeVar, Union, get_args, get_origin

T = TypeVar("T")

MaybeAwaitable = Union[Awaitable[T], T]


def match_type_union_aware(obj_type: Type, match_func: Callable[[Type], bool]) -> Optional[Any]:
    """Check if given obj_type matches match_func and return matched type.

    In case when obj_type is a Union (for e.g. result of Optional[...] makes Union[..., None]),
    match_func wil be run for Union arguments, first match will returned matched argument.
    """
    if match_func(obj_type):
        return obj_type

    if get_origin(obj_type) == Union:
        for union_obj_type in get_args(obj_type):
            if match_func(union_obj_type):
                return union_obj_type

    return None



import asyncio
import inspect
from typing import AsyncIterator, Awaitable, Tuple, TypeVar, Union

X = TypeVar("X")
Y = TypeVar("Y")


class Zip:
    """Merges two async iterators into a single async iterator.

    If value from any of the iterators is an awaitable, yielded value is a
    tuple-returning awaitable. If all values are non-awaitables, yielded value is a tuple.

    If streams have different lengths, there will be as many pairs as there are items in the shorter
    one.

    Sample usage::

        async def str_stream():
            yield "foo"
            yield "bar"

        async def int_stream():
            yield 1
            yield 2

        async for pair in Zip(int_stream())(str_stream()):
            print(pair)
            #   ("foo", 1)
            #   ("bar", 2)

    Or with an awaitable::

        async def get_x():
            return "baz"

        async def str_stream():
            yield get_x()
            yield get_x()

        async def int_stream():
            yield 1
            yield 2

        async for awaitable_pair in Zip(int_stream())(str_stream()):
            print(await awaitable_pair)
            #   ("baz", 1)
            #   ("baz", 2)
    """

    def __init__(self, main_stream: AsyncIterator[Union[X, Awaitable[X]]]):
        self._main_stream = main_stream

    async def __call__(
        self, other_stream: AsyncIterator[Union[Y, Awaitable[Y]]]
    ) -> AsyncIterator[Union[Tuple[Y, X], Awaitable[Tuple[Y, X]]]]:
        async for main_value in self._main_stream:
            try:
                other_value = await other_stream.__anext__()
            except StopAsyncIteration:
                break

            if any(inspect.isawaitable(val) for val in (main_value, other_value)):
                yield self._merge_awaitables(other_value, main_value)  # type: ignore
            else:
                yield other_value, main_value  # type: ignore

    async def _merge_awaitables(
        self, val_1: Union[Y, Awaitable[Y]], val_2: Union[X, Awaitable[X]]
    ) -> Awaitable[Tuple[Y, X]]:
        if inspect.isawaitable(val_1):
            if inspect.isawaitable(val_2):
                values = await asyncio.gather(val_1, val_2)
            else:
                values = (await val_1, val_2)
        elif inspect.isawaitable(val_2):
            values = (val_1, await val_2)
        else:
            values = (val_1, val_2)

        return tuple(values)  # type: ignore



import logging
from typing import Awaitable, Callable, List, Optional, Sequence

from golem.managers.base import ProposalManager, ProposalManagerPlugin
from golem.managers.proposal.plugins.negotiating.negotiating_plugin import NegotiatingPlugin
from golem.node.node import GolemNode
from golem.resources import Proposal
from golem.utils.logging import trace_span


class DefaultProposalManager(ProposalManager):
    def __init__(
        self,
        golem: GolemNode,
        get_initial_proposal: Callable[[], Awaitable[Proposal]],
        plugins: Optional[Sequence[ProposalManagerPlugin]] = None,
    ) -> None:
        self._golem = golem
        self._get_initial_proposal = get_initial_proposal
        self._get_proposal_with_plugins = self._get_initial_proposal
        self._plugins: List[ProposalManagerPlugin] = (
            list(plugins) if plugins is not None else [NegotiatingPlugin()]
        )

    @trace_span("Getting draft proposal", show_results=True, log_level=logging.INFO)
    async def get_draft_proposal(self) -> Proposal:
        return await self._get_proposal_with_plugins()

    @trace_span("Starting DefaultProposalManager", log_level=logging.INFO)
    async def start(self) -> None:
        for p in self._plugins:
            p.set_proposal_callback(self._get_proposal_with_plugins)
            self._get_proposal_with_plugins = p.get_proposal
            await p.start()

    @trace_span("Stopping DefaultProposalManager", log_level=logging.INFO)
    async def stop(self) -> None:
        for p in reversed(self._plugins):
            await p.stop()



import asyncio
import logging
from datetime import timedelta
from typing import Callable, Optional, Sequence

from golem.managers.base import ScorerWithOptionalWeight
from golem.managers.proposal.plugins.buffer import ProposalBuffer
from golem.managers.proposal.plugins.scoring import ProposalScoringMixin
from golem.resources import Proposal
from golem.utils.asyncio import (
    Buffer,
    ExpirableBuffer,
    SimpleBuffer,
    create_task_with_logging,
    ensure_cancelled,
)
from golem.utils.logging import get_trace_id_name, trace_span
from golem.utils.typing import MaybeAwaitable

logger = logging.getLogger(__name__)


class ProposalScoringBuffer(ProposalScoringMixin, ProposalBuffer):
    def __init__(
        self,
        min_size: int,
        max_size: int,
        fill_concurrency_size=1,
        fill_at_start=False,
        get_expiration_func: Optional[
            Callable[[Proposal], MaybeAwaitable[Optional[timedelta]]]
        ] = None,
        on_expiration_func: Optional[Callable[[Proposal], MaybeAwaitable[None]]] = None,
        scoring_debounce: timedelta = timedelta(seconds=10),
        proposal_scorers: Optional[Sequence[ScorerWithOptionalWeight]] = None,
    ) -> None:
        super().__init__(
            min_size=min_size,
            max_size=max_size,
            fill_concurrency_size=fill_concurrency_size,
            fill_at_start=fill_at_start,
            get_expiration_func=None,
            on_expiration_func=on_expiration_func,
            proposal_scorers=proposal_scorers,
        )

        self._scoring_debounce = scoring_debounce

        # Postponing argument would disable expiration from ProposalBuffer parent
        # as we want to expire only scored proposals instead
        self._get_expiration_func = get_expiration_func

        scored_buffer: Buffer[Proposal] = SimpleBuffer()

        if get_expiration_func is not None:
            scored_buffer = ExpirableBuffer(
                buffer=scored_buffer,
                get_expiration_func=get_expiration_func,
                on_expired_func=self._on_expired,
            )

        self._buffer_scored: Buffer[Proposal] = scored_buffer
        self._background_loop_task: Optional[asyncio.Task] = None

    @trace_span()
    async def start(self) -> None:
        await super().start()

        self._background_loop_task = create_task_with_logging(
            self._background_loop(), trace_id=get_trace_id_name(self, "background-loop")
        )

    @trace_span()
    async def stop(self) -> None:
        await super().stop()

        if self._background_loop_task is not None:
            await ensure_cancelled(self._background_loop_task)
            self._background_loop_task = None

    async def _on_added(self, proposal: Proposal) -> None:
        pass  # explicit no-op

    async def _background_loop(self) -> None:
        while True:
            logger.debug(
                "Waiting for any proposals to score with debounce of `%s`...",
                self._scoring_debounce,
            )

            try:
                proposals = await self._buffer.get_requested(self._scoring_debounce)
            except Exception as e:
                await self._buffer_scored.set_exception(e)
                logger.debug(
                    "Encountered unexpected exception while getting proposal,"
                    " exception is set and background loop will be stopped!"
                )

                return

            logger.debug(
                "Waiting for any proposals done, %d new proposals will be scored", len(proposals)
            )

            proposals.extend(await self._buffer_scored.get_all())

            logger.debug("Scoring total %d proposals...", len(proposals))

            scored_proposals = await self.do_scoring(proposals)
            await self._buffer_scored.put_all([proposal for _, proposal in scored_proposals])

            logger.debug("Scoring total %d proposals done", len(proposals))

    async def _get_buffered_proposal(self) -> Proposal:
        return await self._buffer_scored.get()

    def _get_buffered_proposals_count(self) -> int:
        return super()._get_buffered_proposals_count() + self._buffer_scored.size()



import asyncio
import logging
from datetime import timedelta
from decimal import Decimal
from typing import TYPE_CHECKING, Optional, Tuple, Union

from ya_payment import RequestorApi, models

from golem.resources.allocation.allocation import Allocation
from golem.resources.base import _NULL, Resource, api_call_wrapper
from golem.resources.exceptions import PaymentValidationException
from golem.resources.invoice.events import NewInvoice
from golem.resources.utils.infrastructure import InfrastructureProps
from golem.resources.utils.payment import (
    LinearCoeffs,
    PayDocumentStatus,
    eth_decimal,
    validate_payment_max_cost,
)

if TYPE_CHECKING:
    from golem.node import GolemNode
    from golem.resources.agreement import Agreement  # noqa
    from golem.resources.agreement.data import AgreementData  # noqa

logger = logging.getLogger(__name__)


class Invoice(Resource[RequestorApi, models.Invoice, "Agreement", _NULL, _NULL]):
    """A single invoice on the Golem Network.

    Usually created by a :any:`GolemNode` initialized with `collect_payment_events = True`.
    """

    @property
    def agreement(self) -> "Agreement":
        return self.parent

    def __init__(self, node: "GolemNode", id_: str, data: Optional[models.Invoice] = None):
        super().__init__(node, id_, data)
        asyncio.create_task(node.event_bus.emit(NewInvoice(self)))

    async def get_time_and_amount_since_latest_debit_notes(
        self, total_amount: Decimal
    ) -> Tuple[Optional[timedelta], Decimal]:
        """Get cumulative time and amount since last debit notes from all activities."""
        cumulative_time_from_all_activities = timedelta()
        cumulative_amount_from_all_activities = Decimal(0)

        for activity in self.agreement.activities:
            # Look for a newest Debit Note that came before the invoice
            for debit_note in sorted(
                activity.debit_notes, key=lambda dn: dn.created_at, reverse=True
            ):
                await debit_note.get_data()  # ensure there is access to `debit_note.data``
                if debit_note.created_at < self.created_at:
                    cumulative_time_from_all_activities += self.created_at - debit_note.created_at
                    cumulative_amount_from_all_activities += eth_decimal(
                        debit_note.data.total_amount_due
                    )
                break

        if cumulative_time_from_all_activities == timedelta():
            cumulative_time_from_all_activities = None  # type: ignore[assignment]

        return (
            cumulative_time_from_all_activities,
            total_amount - cumulative_amount_from_all_activities,
        )

    async def validate_and_accept(self, allocation: Allocation) -> None:
        """Validate invoice and accept using a given :any:`Allocation`."""
        invoice_data = await self.get_data(force=True)
        if invoice_data.status != PayDocumentStatus.RECEIVED:
            logger.warning(f"Wrong status of invoice {invoice_data.status} != RECEIVED")
            return

        if self.agreement.approved_at is None:
            logger.warning("Agreement was not approved")
            return

        try:
            agreement_data: "AgreementData" = await self.agreement.get_agreement_data()
            amount_due = eth_decimal(invoice_data.amount)
            (
                cumulative_time_since_last_dn,
                cumulative_amount_since_last_dn,
            ) = await self.get_time_and_amount_since_latest_debit_notes(amount_due)
            coeffs = LinearCoeffs.from_properties(agreement_data.properties)
            infrastructure = InfrastructureProps.from_properties(agreement_data.properties)

            max_cost, max_cost_since_latest_debit_notes = validate_payment_max_cost(
                coeffs=coeffs,
                inf=infrastructure,
                duration=agreement_data.agreement_duration,
                amount=amount_due,
                time_since_last_debit_note=cumulative_time_since_last_dn,
                amount_since_last_debit_note=cumulative_amount_since_last_dn,
            )
        except PaymentValidationException:
            logger.warning(
                f"Invoice {self.id} validation failed. {coeffs=}, {infrastructure=}, "
                f"{agreement_data.agreement_duration=}, {amount_due=}, "
                f"{cumulative_time_since_last_dn=}, {cumulative_amount_since_last_dn=}",
                exc_info=True,
            )
            return

        logger.info(
            f"Accepting Invoice for {self.agreement.id=} "
            f"{amount_due=} {max_cost=} {max_cost_since_latest_debit_notes=}"
        )
        await self.accept(allocation, amount_due)

    async def accept_full(self, allocation: Allocation) -> None:
        """Accept full invoice amount using a given :any:`Allocation`."""
        amount_str = (await self.get_data()).amount
        await self.accept(allocation, Decimal(amount_str))

    @api_call_wrapper(retry_count=5)
    async def accept(self, allocation: Allocation, amount: Union[Decimal, float]) -> None:
        acceptance = models.Acceptance(
            total_amount_accepted=str(amount), allocation_id=allocation.id
        )
        await self.api.accept_invoice(self.id, acceptance)



from dataclasses import dataclass, field, replace
from datetime import datetime, timedelta, timezone
from decimal import Decimal
from os import getenv
from typing import Optional, Tuple

from golem.payload.base import Payload, constraint, prop
from golem.payload.constraints import Constraints
from golem.payload.properties import Properties

PROP_RUNTIME_NAME = "golem.runtime.name"
PROP_RUNTIME_CAPABILITIES = "golem.runtime.capabilities"
PROP_INF_CPU_THREADS = "golem.inf.cpu.threads"
PROP_INF_MEM = "golem.inf.mem.gib"
PROP_INF_STORAGE = "golem.inf.storage.gib"

PROP_DEBIT_NOTES_ACCEPT_TIMEOUT = "golem.com.payment.debit-notes.accept-timeout?"

PROP_DEBIT_NOTES_INTERVAL = "golem.com.scheme.payu.debit-note.interval-sec?"
PROP_PAYMENT_TIMEOUT = "golem.com.scheme.payu.payment-timeout-sec?"

PROP_PRICING_MODEL = "golem.com.pricing.model"
PROP_PRICING_LINEAR_COEFFS = "golem.com.pricing.model.linear.coeffs"
PROP_USAGE_VECTOR = "golem.com.usage.vector"

DEFAULT_PAYMENT_DRIVER: str = getenv("YAGNA_PAYMENT_DRIVER", "erc20").lower()
DEFAULT_PAYMENT_NETWORK: str = getenv("YAGNA_PAYMENT_NETWORK", "goerli").lower()

DEFAULT_LIFETIME = timedelta(minutes=30)
DEFAULT_SUBNET: str = getenv("YAGNA_SUBNET", "public")


@dataclass
class NodeInfo(Payload):
    """Properties and constraints describing the information regarding the node."""

    name: Optional[str] = prop("golem.node.id.name", default=None)
    """human-readable name of the Golem node"""

    subnet_tag: Optional[str] = prop("golem.node.debug.subnet", default=DEFAULT_SUBNET)
    _subnet_tag_constraint: Optional[str] = constraint(
        "golem.node.debug.subnet", default=None, init=False
    )
    """the name of the subnet within which the Demands and Offers are matched"""

    def __post_init__(self):
        self._subnet_tag_constraint = self.subnet_tag


@dataclass
class ActivityInfo(Payload):
    """Activity-related Properties."""

    cost_cap: Optional[Decimal] = prop("golem.activity.cost_cap", default=None)
    """Sets a Hard cap on total cost of the Activity (regardless of the usage vector or
    pricing function). The Provider is entitled to 'kill' an Activity which exceeds the
    capped cost amount indicated by Requestor.
    """

    cost_warning: Optional[Decimal] = prop("golem.activity.cost_warning", default=None)
    """Sets a Soft cap on total cost of the Activity (regardless of the usage vector or
    pricing function). When the cost_warning amount is reached for the Activity,
    the Provider is expected to send a Debit Note to the Requestor, indicating
    the current amount due
    """

    timeout_secs: Optional[float] = prop("golem.activity.timeout_secs", default=None)
    """A timeout value for batch computation (eg. used for container-based batch
    processes). This property allows to set the timeout to be applied by the Provider
    when running a batch computation: the Requestor expects the Activity to take
    no longer than the specified timeout value - which implies that
    eg. the golem.usage.duration_sec counter shall not exceed the specified
    timeout value.
    """

    expiration: Optional[datetime] = prop("golem.srv.comp.expiration", default=None)
    """The datetime until which any started activities will last."""

    lifetime: Optional[timedelta] = field(default=None)
    """Lifetime of the activities from the moment the demand is placed.

    Convenience property, used only in case the actual `expiration` is not provided.
    """

    multi_activity: Optional[bool] = prop("golem.srv.caps.multi-activity", default=None)
    """Whether client supports multi_activity (executing more than one activity per agreement).
    """

    def __post_init__(self):
        if self.expiration and self.lifetime:
            raise ValueError(
                "Ambiguous definition - either the expiration or the lifetime must be provided."
            )

        # in case the expiration itself is not provided, set the default lifetime instead
        # so that the expiration can be dynamically constructed when needed
        if not self.expiration and not self.lifetime:
            self.lifetime = DEFAULT_LIFETIME

    async def build_properties_and_constraints(self) -> Tuple[Properties, Constraints]:
        # we don't want to freeze the expiration, so we're making a copy here
        if not self.expiration:
            activity_info = replace(self)
            assert activity_info.lifetime  # set in `__post_init__`
            activity_info.expiration = datetime.now(timezone.utc) + activity_info.lifetime
            return await activity_info.build_properties_and_constraints()

        return await super().build_properties_and_constraints()


@dataclass
class PaymentInfo(Payload):
    chosen_payment_platform: Optional[str] = prop("golem.com.payment.chosen-platform", default=None)
    """Payment platform selected to be used for this demand."""

    debit_notes_accept_timeout: int = prop(PROP_DEBIT_NOTES_ACCEPT_TIMEOUT, default=120)

    debit_notes_interval: Optional[int] = prop(PROP_DEBIT_NOTES_INTERVAL, default=None)
    payment_timeout: Optional[int] = prop(PROP_PAYMENT_TIMEOUT, default=None)



from typing import TYPE_CHECKING, Optional

from golem.resources.exceptions import ResourceException

if TYPE_CHECKING:
    from golem.resources import ProposalId


class ProposalException(ResourceException):
    pass


class ProposalRejected(ResourceException):
    def __init__(self, proposal_id: Optional["ProposalId"], reason: str):
        self._proposal_id = proposal_id
        self._reason = reason

        super().__init__(f"Proposal `{proposal_id}` rejected! Reason: `{reason}`")

    @property
    def proposal_id(self) -> Optional["ProposalId"]:
        return self._proposal_id

    @property
    def reason(self) -> str:
        return self._reason



import aiohttp

# TODO: Implement Port/Adapter pattern and make aiohttp dependency optional for http requests for
#   more decoupling


async def make_http_get_request(url: str, raise_exceptions: bool = False) -> str:
    async with aiohttp.ClientSession() as client:
        response = await client.get(url)

        if response.status != 200 and raise_exceptions:
            response.raise_for_status()

        return await response.text()


async def make_http_head_request(url: str, raise_exceptions: bool = False) -> None:
    async with aiohttp.ClientSession() as client:
        response = await client.head(url, allow_redirects=True)

        if response.status != 200 and raise_exceptions:
            response.raise_for_status()



import asyncio
from collections import defaultdict
from datetime import datetime, timezone
from decimal import Decimal
from typing import Any, DefaultDict, Dict, Iterable, List, Optional, Set, Type, Union
from uuid import uuid4

from golem.event_bus import EventBus
from golem.event_bus.in_memory import InMemoryEventBus
from golem.node.events import SessionStarted, ShutdownFinished, ShutdownStarted
from golem.payload import Payload
from golem.payload import defaults as payload_defaults
from golem.resources import (
    Activity,
    Agreement,
    Allocation,
    DebitNote,
    DebitNoteEventCollector,
    Demand,
    DemandBuilder,
    Invoice,
    InvoiceEventCollector,
    Network,
    PoolingBatch,
    Proposal,
    Resource,
    TResource,
)
from golem.utils.low import ApiConfig, ApiFactory


class _RandomSessionId:
    pass


class GolemNode:
    """Main entrypoint to the python Golem API, communicates with `yagna`.

    GolemNode object corresponds to a single identity and a single running `yagna` instance
    (--> it's identified by a (APPKEY, YAGNA_URL) pair) and can operate on different
    subnets / networks. Multiple GolemNode instances can be used to access different
    identities / `yagna` instances.

    Usage::

        golem = GolemNode()
        async with golem:
            #   Interact with the Golem Network

    """

    def __init__(
        self,
        app_key: Optional[str] = None,
        *,
        base_url: Optional[str] = None,
        collect_payment_events: bool = True,
        app_session_id: Optional[Union[str, Type[_RandomSessionId]]] = _RandomSessionId,
    ):
        """Init GolemNode.

        :param app_key: App key used as an authentication token for all `yagna` calls.
                        Defaults to the `YAGNA_APPKEY` env variable.
        :param base_url: Base url for all `yagna` APIs. Defaults to `YAGNA_API_URL` env
                         variable or http://127.0.0.1:7465.
        :param collect_payment_events: If True, GolemNode will watch for incoming
            debit notes/invoices and create corresponding objects (--> :any:`NewResource` events
            will be emitted).
        :param app_session_id: A correlation/session identifier. :any:`GolemNode` objects with the
            same `app_session_id` will receive the same debit note/invoice/agreement events.
            Defaults to a random sting. If set to `None`, this GolemNode will receive all events
            regardless of their corresponding session ids.
        """
        config_kwargs = {
            param: value
            for param, value in {"app_key": app_key, "base_url": base_url}.items()
            if value is not None
        }
        self._api_config = ApiConfig(**config_kwargs)
        self._collect_payment_events = collect_payment_events
        self.app_session_id = uuid4().hex if app_session_id is _RandomSessionId else app_session_id

        #   All created Resources will be stored here
        #   (This is done internally by the metaclass of the Resource)
        self._resources: DefaultDict[Type[Resource], Dict[str, Resource]] = defaultdict(dict)
        self._autoclose_resources: Set[Resource] = set()
        self._event_bus = InMemoryEventBus()

        self._invoice_event_collector = InvoiceEventCollector(self)
        self._debit_note_event_collector = DebitNoteEventCollector(self)

    @property
    def app_key(self) -> str:
        return self._api_config.app_key

    ########################
    #   Start/stop interface
    async def __aenter__(self) -> "GolemNode":
        """Start. Initialize all the APIs and the event bus."""
        await self.start()
        return self

    async def __aexit__(self, *exc_info: Any) -> None:
        """Shutdown.

        Stop collecting yagna events, close all resources created with autoclose=True, close
        APIs etc.
        """
        await self.aclose()

    async def start(self) -> None:
        await self.event_bus.start()

        api_factory = ApiFactory(self._api_config)
        self._ya_market_api = api_factory.create_market_api_client()
        self._ya_activity_api = api_factory.create_activity_api_client()
        self._ya_payment_api = api_factory.create_payment_api_client()
        self._ya_net_api = api_factory.create_net_api_client()

        if self._collect_payment_events:
            self._invoice_event_collector.start_collecting_events()
            self._debit_note_event_collector.start_collecting_events()

        await self.event_bus.emit(SessionStarted(self))

    async def aclose(self) -> None:
        await self.event_bus.emit(ShutdownStarted(self))
        self._set_no_more_children()
        self._stop_event_collectors()
        await self._close_autoclose_resources()
        await self._close_apis()
        await self.event_bus.emit(ShutdownFinished(self))

        await self.event_bus.stop()

    def _stop_event_collectors(self) -> None:
        demands = self.all_resources(Demand)
        batches = self.all_resources(PoolingBatch)
        payment_event_collectors = [self._invoice_event_collector, self._debit_note_event_collector]

        for event_collector in demands + batches + payment_event_collectors:
            event_collector.stop_collecting_events()

    def _set_no_more_children(self) -> None:
        for resources in self._resources.values():
            for resource in resources.values():
                resource.set_no_more_children()

    async def _close_apis(self) -> None:
        await asyncio.gather(
            self._ya_market_api.close(),
            self._ya_activity_api.close(),
            self._ya_payment_api.close(),
            self._ya_net_api.close(),
        )

    async def _close_autoclose_resources(self) -> None:
        agreement_msg = "Work finished"
        activity_tasks = [r.destroy() for r in self._autoclose_resources if isinstance(r, Activity)]
        agreement_tasks = [
            r.terminate(agreement_msg)
            for r in self._autoclose_resources
            if isinstance(r, Agreement)
        ]
        demand_tasks = [r.unsubscribe() for r in self._autoclose_resources if isinstance(r, Demand)]
        allocation_tasks = [
            r.release() for r in self._autoclose_resources if isinstance(r, Allocation)
        ]
        network_tasks = [r.remove() for r in self._autoclose_resources if isinstance(r, Network)]
        if activity_tasks:
            await asyncio.gather(*activity_tasks)
        if agreement_tasks:
            await asyncio.gather(*agreement_tasks)
        if demand_tasks:
            await asyncio.gather(*demand_tasks)
        if allocation_tasks:
            await asyncio.gather(*allocation_tasks)
        if network_tasks:
            await asyncio.gather(*network_tasks)

    ###########################
    #   Create new resources
    async def create_allocation(
        self,
        amount: Union[Decimal, float],
        network: str = payload_defaults.DEFAULT_PAYMENT_NETWORK,
        driver: str = payload_defaults.DEFAULT_PAYMENT_DRIVER,
        autoclose: bool = True,
    ) -> Allocation:
        """Create a new allocation.

        :param amount: Amount of GLMs to be allocated
        :param network: Payment network
        :param driver: Payment driver
        :param autoclose: Release allocation on :func:`__aexit__`
        """
        decimal_amount = Decimal(amount)

        #   TODO (?): https://github.com/golemfactory/golem-core-python/issues/34
        allocation = await Allocation.create_any_account(self, decimal_amount, network, driver)
        if autoclose:
            self.add_autoclose_resource(allocation)
        return allocation

    async def create_demand(
        self,
        payload: Payload,
        subnet: Optional[str] = payload_defaults.DEFAULT_SUBNET,
        expiration: Optional[datetime] = None,
        allocations: Iterable[Allocation] = (),
        autoclose: bool = True,
        autostart: bool = True,
    ) -> Demand:
        """Subscribe a new demand.

        :param payload: Details of the demand
        :param subnet: Subnet tag
        :param expiration: Timestamp when all agreements based on this demand will expire
            TODO: is this correct?
        :param allocations: Allocations that will be included in the description of this demand.
        :param autoclose: Unsubscribe demand on :func:`__aexit__`
        :param autostart: Immediately start collecting yagna events for this :any:`Demand`.
            Without autostart events for this demand will start being collected after a call to
            :func:`Demand.start_collecting_events`.
        """
        if expiration is None:
            expiration = datetime.now(timezone.utc) + payload_defaults.DEFAULT_LIFETIME

        builder = DemandBuilder()
        await builder.add(payload_defaults.ActivityInfo(expiration=expiration, multi_activity=True))
        await builder.add(payload_defaults.NodeInfo(subnet_tag=subnet))

        await builder.add(payload)
        await self._add_builder_allocations(builder, allocations)

        demand = await Demand.create_from_properties_constraints(
            self, builder.properties, builder.constraints
        )

        if autostart:
            demand.start_collecting_events()
        if autoclose:
            self.add_autoclose_resource(demand)
        return demand

    async def create_network(
        self,
        ip: str,
        mask: Optional[str] = None,
        gateway: Optional[str] = None,
        autoclose: bool = True,
        add_requestor: bool = True,
        requestor_ip: Optional[str] = None,
    ) -> Network:
        """Create a new :any:`Network`.

        :param ip: IP address of the network. May contain netmask, e.g. `192.168.0.0/24`.
        :param mask: Optional netmask (only if not provided within the `ip` argument).
        :param gateway: Optional gateway address for the network.
        :param autoclose: Remove network on :func:`__aexit__`
        :param add_requestor: If True, adds requestor with ip `requestor_ip` to the network.
            If False, requestor will be able to interact with other nodes only
            after an additional call to :func:`add_to_network`.
        :param requestor_ip: Ip of the requestor node in the network. Ignored if not
            `add_requestor`. If `None`, next free ip will be assigned.
        """
        network = await Network.create(self, ip, mask, gateway)
        if autoclose:
            self.add_autoclose_resource(network)
        if add_requestor:
            await self.add_to_network(network, requestor_ip)
        return network

    async def _add_builder_allocations(
        self, builder: DemandBuilder, allocations: Iterable[Allocation]
    ) -> None:
        # TODO (?): https://github.com/golemfactory/golem-core-python/issues/35

        for allocation in allocations:
            await builder.add(await allocation.get_demand_spec())

    ###########################
    #   Single-resource factories for already existing resources
    def allocation(self, allocation_id: str) -> Allocation:
        """Return an :any:`Allocation` with a given id (assumed to be correct, there is no \
        validation)."""
        return Allocation(self, allocation_id)

    def debit_note(self, debit_note_id: str) -> DebitNote:
        """Return an :any:`DebitNote` with a given id (assumed to be correct, there is no \
        validation)."""
        return DebitNote(self, debit_note_id)

    def invoice(self, invoice_id: str) -> Invoice:
        """Return an :any:`Invoice` with a given id (assumed to be correct, there is no \
        validation)."""
        return Invoice(self, invoice_id)

    def demand(self, demand_id: str) -> Demand:
        """Return a :any:`Demand` with a given id (assumed to be correct, there is no \
        validation)."""
        return Demand(self, demand_id)

    def proposal(self, proposal_id: str, demand_id: str) -> Proposal:
        """Return a :any:`Proposal` with a given id (assumed to be correct, there is no validation).

        Id of a proposal has a meaning only in the context of a demand,
        so demand_id is also necessary (and also not validated).
        """
        demand = self.demand(demand_id)
        return demand.proposal(proposal_id)

    def agreement(self, agreement_id: str) -> Agreement:
        """Return an :any:`Agreement` with a given id (assumed to be correct, there is no \
        validation)."""
        return Agreement(self, agreement_id)

    def activity(self, activity_id: str) -> Activity:
        """Return an :any:`Activity` with a given id (assumed to be correct, there is no \
        validation)."""
        return Activity(self, activity_id)

    def batch(self, batch_id: str, activity_id: str) -> PoolingBatch:
        """Return a :any:`PoolingBatch` with a given id (assumed to be correct, there is no \
        validation).

        Id of a batch has a meaning only in the context of an activity,
        so activity_id is also necessary (and also not validated).
        """
        activity = self.activity(activity_id)
        return activity.batch(batch_id)

    ##########################
    #   Multi-resource factories for already existing resources
    async def allocations(self) -> List[Allocation]:
        """Return a list of :any:`Allocation` objects corresponding to all current allocations.

        These are all allocations related to the current APP_KEY - it doesn't matter if they were
        created with this :class:`GolemNode` instance (or if :class:`GolemNode` was used at all).
        """
        return await Allocation.get_all(self)

    async def demands(self) -> List[Demand]:
        """Return a list of :any:`Demand` objects corresponding to all current demands.

        These are all demands subscribed with the current APP_KEY - it doesn't matter if they were
        created with this :class:`GolemNode` instance (or if :class:`GolemNode` was used at all).
        """
        return await Demand.get_all(self)

    async def invoices(self) -> List[Invoice]:
        """Return a list of :any:`Invoice` objects corresponding to all invoices received by this \
        node."""
        return await Invoice.get_all(self)

    async def debit_notes(self) -> List[DebitNote]:
        """Return a list of :any:`DebitNote` objects corresponding to all debit notes received by \
        this node."""
        return await DebitNote.get_all(self)

    async def networks(self) -> List[Network]:
        """Return a list of :any:`Network` objects corresponding to all networks created by this \
        node.

        These are all networks created with the current APP_KEY - it doesn't matter if they were
        created with this :class:`GolemNode` instance (or if :class:`GolemNode` was used at all).
        """
        return await Network.get_all(self)

    ##########################
    #   Events
    @property
    def event_bus(self) -> EventBus:
        """Return the :any:`EventBus` used by this :class:`GolemNode`.

        Any :any:`Event` triggered by this :class:`GolemNode` or any related object
        will be sent there and passed to registered listeners.
        """
        return self._event_bus

    #########
    #   Other
    async def add_to_network(self, network: Network, ip: Optional[str] = None) -> None:
        """Add requestor to the network.

        :param network: A :any:`Network` we're adding the requestor to.
        :param ip: IP of the requestor node, defaults to a new free IP in the network.

        This is only necessary if we either called :func:`create_network` with
        `add_requestor=False`, or we want the requestor to have multiple IPs in the network
        (TODO: is there a scenario where this makes sense?).
        """
        await network.add_requestor_ip(ip)

    def add_autoclose_resource(
        self, resource: Union["Allocation", "Demand", "Agreement", "Activity", "Network"]
    ) -> None:
        self._autoclose_resources.add(resource)

    def all_resources(self, cls: Type[TResource]) -> List[TResource]:
        """Return all known resources of a given type."""
        return list(self._resources[cls].values())  # type: ignore

    def __str__(self) -> str:
        lines = [
            f"{type(self).__name__}(",
            f"  app_key = {self.app_key},",
            f"  app_session_id = {self.app_session_id},",
            f"  market_url = {self._api_config.market_url},",
            f"  payment_url = {self._api_config.payment_url},",
            f"  activity_url = {self._api_config.activity_url},",
            f"  net_url = {self._api_config.net_url},",
            ")",
        ]
        return "\n".join(lines)



from golem.node.events import GolemNodeEvent, SessionStarted, ShutdownFinished, ShutdownStarted
from golem.node.node import GolemNode

__all__ = (
    "GolemNode",
    "GolemNodeEvent",
    "SessionStarted",
    "ShutdownStarted",
    "ShutdownFinished",
)



import logging
from datetime import timedelta
from decimal import Decimal
from typing import Optional

from golem.payload import defaults
from golem.resources import LinearCoeffs, ProposalData

logger = logging.getLogger(__name__)


class LinearAverageCostPricing:
    def __init__(self, average_cpu_load: float, average_duration: timedelta) -> None:
        self._average_cpu_load = average_cpu_load
        self._average_duration = average_duration

    def __call__(self, proposal_data: ProposalData) -> Optional[float]:
        coeffs = LinearCoeffs.from_properties(proposal_data.properties)

        if coeffs is None:
            return None

        return self._calculate_cost(coeffs)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"average_cpu_load={self._average_cpu_load}, "
            f"average_duration={self._average_duration})"
        )

    def _calculate_cost(self, coeffs: LinearCoeffs) -> float:
        average_duration_sec = Decimal(self._average_duration.total_seconds())

        average_initial_price = coeffs.price_initial
        average_duration_cost = coeffs.price_duration_sec * average_duration_sec
        average_cpu_cost = (
            coeffs.price_cpu_sec * Decimal(self._average_cpu_load) * average_duration_sec
        )

        return float(average_initial_price + average_duration_cost + average_cpu_cost)


class LinearPerCpuAverageCostPricing(LinearAverageCostPricing):
    def __call__(self, proposal_data: ProposalData) -> Optional[float]:
        coeffs = LinearCoeffs.from_properties(proposal_data.properties)
        cpu_count = proposal_data.properties.get(defaults.PROP_INF_CPU_THREADS)

        if coeffs is None or not cpu_count:
            return None

        cpu_count = Decimal(cpu_count)

        coeffs.price_initial /= cpu_count
        coeffs.price_duration_sec /= cpu_count

        return self._calculate_cost(coeffs)



import asyncio
import logging
from typing import Awaitable, Callable, MutableMapping, Sequence

from golem.managers import DemandManager
from golem.node import GolemNode
from golem.resources import Proposal
from golem.utils.asyncio import (
    Buffer,
    SimpleBuffer,
    create_task_with_logging,
    ensure_cancelled_many,
)
from golem.utils.logging import get_trace_id_name, trace_span


class AggregatingDemandManager(DemandManager):
    """DemandManager that combines multiple get_initial_proposal functions into one."""

    def __init__(
        self,
        golem: GolemNode,
        get_initial_proposal_funcs: Sequence[Callable[[], Awaitable[Proposal]]],
    ) -> None:
        self._golem = golem
        self._get_initial_proposal_funcs = get_initial_proposal_funcs

        self._lock = asyncio.Lock()

        self._task_map: MutableMapping[Callable[[], Awaitable[Proposal]], asyncio.Task] = {}

        self._buffer: Buffer[Proposal] = SimpleBuffer()

    @trace_span("Stopping AggregatingDemandManager", log_level=logging.INFO)
    async def stop(self) -> None:
        await ensure_cancelled_many(self._task_map.values())
        self._task_map.clear()

    @trace_span("Getting initial proposal", show_results=True)
    async def get_initial_proposal(self) -> Proposal:
        """Call and return first completed Proposal from `get_initial_proposal_funcs`.

        `get_initial_proposal_funcs` will be called concurrently, where not returned results will
        be saved and returned with the next call. If no proposals are saved, and some other
        functions are pending, completed functions will be run again. In case of multiple proposals
        are saved, they will be returned in completion  order.
        """

        if self._buffer.size():
            return await self._buffer.get()

        async with self._lock:
            for idx, func in enumerate(self._get_initial_proposal_funcs):
                if func not in self._task_map:
                    self._task_map[func] = create_task_with_logging(
                        self._feed_queue(func),
                        trace_id=get_trace_id_name(self, f"feed-func-{idx}"),
                    )

        return await self._buffer.get()

    async def _feed_queue(self, func: Callable[[], Awaitable[Proposal]]):
        try:
            proposal = await func()
        except asyncio.CancelledError:
            # we don't want to store cancellation error
            raise
        except Exception as e:
            await self._buffer.set_exception(e)
        else:
            self._buffer.reset_exception()
            await self._buffer.put(proposal)

        async with self._lock:
            del self._task_map[func]



import asyncio
import logging
from typing import Awaitable, Callable

from golem.managers.activity.mixins import ActivityPrepareReleaseMixin, ActivityWrapper
from golem.managers.base import ActivityManager
from golem.managers.mixins import BackgroundLoopMixin
from golem.node import GolemNode
from golem.resources import Activity, Agreement
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)


@Activity.register
class PoolActivity(ActivityWrapper):
    def __init__(self, activity, put_activity_to_pool_func) -> None:
        super().__init__(activity)
        self._put_activity_to_pool_func = put_activity_to_pool_func

    async def destroy(self) -> None:
        await self._put_activity_to_pool_func(self._activity)


class PoolActivityManager(BackgroundLoopMixin, ActivityPrepareReleaseMixin, ActivityManager):
    def __init__(
        self,
        golem: GolemNode,
        get_agreement: Callable[[], Awaitable[Agreement]],
        pool_size: int,
        *args,
        **kwargs,
    ):
        self._get_agreement = get_agreement
        self._event_bus = golem.event_bus

        self._pool_target_size = pool_size
        self._pool_current_size = 0
        self._pool: asyncio.Queue[Activity] = asyncio.Queue()
        super().__init__(*args, **kwargs)

    async def _background_loop(self):
        try:
            while True:
                if self._pool_current_size > self._pool_target_size:
                    # TODO check tasks results and add fallback
                    await asyncio.gather(
                        *[
                            self._release_activity_and_pop_from_pool()
                            for _ in range(self._pool_current_size - self._pool_target_size)
                        ]
                    )
                elif self._pool_current_size < self._pool_target_size:
                    # TODO check tasks results and add fallback
                    await asyncio.gather(
                        *[
                            self._prepare_activity_and_put_to_pool()
                            for _ in range(self._pool_target_size - self._pool_current_size)
                        ]
                    )
                # TODO: Use events instead of sleep
                await asyncio.sleep(0.01)
        finally:
            logger.info(f"Releasing all {self._pool_current_size} activities from the pool")
            # TODO cancel release adn prepare tasks
            await asyncio.gather(
                *[
                    self._release_activity_and_pop_from_pool()
                    for _ in range(self._pool_current_size)
                ]
            )

    @trace_span()
    async def _release_activity_and_pop_from_pool(self):
        activity = await self._pool.get()
        self._pool.task_done()
        self._pool_current_size -= 1
        await self._release_activity(activity)

    @trace_span()
    async def _prepare_activity_and_put_to_pool(self):
        agreement = await self._get_agreement()
        activity = await self._prepare_activity(agreement)
        await self._pool.put(activity)
        self._pool_current_size += 1

    async def _get_activity_from_pool(self):
        activity = await self._pool.get()
        self._pool.task_done()
        logger.debug(f"Activity `{activity}` taken from the pool")
        return activity

    async def _put_activity_to_pool(self, activity):
        await self._pool.put(activity)
        logger.debug(f"Activity `{activity}` back to the pool")

    @trace_span(show_arguments=True, show_results=True)
    async def get_activity(self) -> Activity:
        activity = await self._get_activity_from_pool()
        # mypy doesn't support `ABCMeta.register` https://github.com/python/mypy/issues/2922
        return PoolActivity(activity, self._put_activity_to_pool)  # type: ignore[return-value]



from typing import AsyncIterator, TypeVar

X = TypeVar("X")


class Limit:
    """Limits the number of elements returned by an async generator.

    Sample usage::

        async def yield_thrice():
            yield "foo"
            yield "bar"
            yield "baz"

        # only "foo" and "bar" will be printed
        async for x in Chain(
            yield_thrice(),
            Limit(2),
        ):
            print(x)
    """

    def __init__(self, max_items: int = 1):
        """Init Limit.

        :param max_items: Maximum number of items yielded
        """
        self.max_items = max_items

    async def __call__(self, stream: AsyncIterator[X]) -> AsyncIterator[X]:
        cnt = 0
        async for val in stream:
            yield val
            cnt += 1
            if cnt == self.max_items:
                break



from typing import TYPE_CHECKING, Optional

from golem.resources.exceptions import ResourceException

if TYPE_CHECKING:
    from golem.resources.pooling_batch.pooling_batch import PoolingBatch


class PoolingBatchException(ResourceException):
    pass


class BatchError(PoolingBatchException):
    """Unspecified exception related to the execution of a batch."""

    def __init__(self, batch: "PoolingBatch", msg: Optional[str] = None):
        self._batch = batch

        if msg is None:
            if batch.events:
                event = batch.events[-1]
                msg = f"{batch} failed on command {event.index}: {event.message}"
            else:
                msg = f"{batch} failed to collect any events"

        super().__init__(msg)

    @property
    def batch(self) -> "PoolingBatch":
        return self._batch


class CommandFailed(BatchError):
    """Raised when awaiting for a result of a command that failed."""

    def __init__(self, batch: "PoolingBatch"):
        event = batch.events[-1]
        msg = f"Command {event.index} in batch {batch} failed: {event.message}"
        super().__init__(batch, msg)


class CommandCancelled(BatchError):
    """Raised when awaiting for a result of a command that was not executed at all because a \
    previous command failed."""

    def __init__(self, batch: "PoolingBatch"):
        event = batch.events[-1]
        msg = (
            f"Command was cancelled because {batch} failed before the command was executed. "
            f"Details: command {event.index} failed with {event.message}"
        )
        super().__init__(batch, msg)


class BatchTimeoutError(PoolingBatchException):
    """Raised in :any:`PoolingBatch.wait()` when the batch execution timed out."""

    def __init__(self, batch: "PoolingBatch", timeout: float):
        self._batch = batch
        self._timeout = timeout

        msg = f"{batch} did not finish in {timeout} seconds"
        super().__init__(msg)

    @property
    def batch(self) -> "PoolingBatch":
        return self._batch

    @property
    def timeout(self) -> float:
        return self._timeout



import os
from dataclasses import dataclass, field
from functools import partial
from typing import TYPE_CHECKING, Any, Final, Optional, Type, TypeVar, get_args, no_type_check

import ya_activity
import ya_market
import ya_net
import ya_payment

from golem.exceptions import MissingConfiguration

if TYPE_CHECKING:
    from golem.node import GolemNode
    from golem.resources import Resource


TRequestorApi = TypeVar("TRequestorApi")


DEFAULT_YAGNA_API_URL: Final[str] = "http://127.0.0.1:7465"


@dataclass
class ApiConfig:
    """Yagna low level API configuration.

    Attributes:
        app_key: Yagna application key.
            If not provided, the default is to get the value from `YAGNA_APPKEY` environment
            variable.
            If no value will be found MissingConfiguration error will be thrown
        api_url: base URL or all REST API URLs. Example value: http://127.0.10.10:7500
            (no trailing slash).
            Uses YAGNA_API_URL environment variable
        market_url: If not provided `api_url` will be used to construct it.
            Uses YAGNA_MARKET_URL environment variable
        payment_url: If not provided `api_url` will be used to construct it.
            Uses YAGNA_PAYMENT_URL environment variable
        net_url: Uses If not provided `api_url` will be used to construct it.
            YAGNA_NET_URL environment variable
        activity_url: If not provided `api_url` will be used to construct it.
            Uses YAGNA_ACTIVITY_URL environment variable
    """

    app_key: str = field(
        default_factory=partial(os.getenv, "YAGNA_APPKEY")
    )  # type: ignore[assignment]
    api_url: str = field(
        default_factory=partial(os.getenv, "YAGNA_API_URL", DEFAULT_YAGNA_API_URL)
    )  # type: ignore[assignment]
    market_url: str = field(
        default_factory=partial(os.getenv, "YAGNA_MARKET_URL")
    )  # type: ignore[assignment]
    payment_url: str = field(
        default_factory=partial(os.getenv, "YAGNA_PAYMENT_URL")
    )  # type: ignore[assignment]
    net_url: str = field(
        default_factory=partial(os.getenv, "YAGNA_NET_URL")
    )  # type: ignore[assignment]
    activity_url: str = field(
        default_factory=partial(os.getenv, "YAGNA_ACTIVITY_URL")
    )  # type: ignore[assignment]

    def __post_init__(self) -> None:
        if self.app_key is None:
            raise MissingConfiguration(key="YAGNA_APPKEY", description="API authentication token")
        self.market_url: str = self.__resolve_url(self.market_url, "/market-api/v1")
        self.payment_url: str = self.__resolve_url(self.payment_url, "/payment-api/v1")
        self.activity_url: str = self.__resolve_url(self.activity_url, "/activity-api/v1")
        self.net_url: str = self.__resolve_url(self.net_url, "/net-api/v1")

    def __resolve_url(self, given_url: Optional[str], prefix: str) -> str:
        return given_url or f"{self.api_url}{prefix}"


class ApiFactory(object):
    """
    REST API's setup and top-level access utility.

    By default, it expects the yagna daemon to be available locally and listening on the
    default port. The urls for the specific APIs are then based on this default base URL.
    """

    def __init__(
        self,
        api_config: ApiConfig,
    ):
        self.__api_config: ApiConfig = api_config

    def create_market_api_client(self) -> ya_market.ApiClient:
        """Return a REST client for the Market API."""
        cfg = ya_market.Configuration(host=self.__api_config.market_url)
        return ya_market.ApiClient(
            configuration=cfg,
            header_name="authorization",
            header_value=f"Bearer {self.__api_config.app_key}",
        )

    def create_payment_api_client(self) -> ya_payment.ApiClient:
        """Return a REST client for the Payment API."""
        cfg = ya_payment.Configuration(host=self.__api_config.payment_url)
        return ya_payment.ApiClient(
            configuration=cfg,
            header_name="authorization",
            header_value=f"Bearer {self.__api_config.app_key}",
        )

    def create_activity_api_client(self) -> ya_activity.ApiClient:
        """Return a REST client for the Activity API."""
        cfg = ya_activity.Configuration(host=self.__api_config.activity_url)
        return ya_activity.ApiClient(
            configuration=cfg,
            header_name="authorization",
            header_value=f"Bearer {self.__api_config.app_key}",
        )

    def create_net_api_client(self) -> ya_net.ApiClient:
        """Return a REST client for the Net API."""
        cfg = ya_net.Configuration(host=self.__api_config.net_url)
        return ya_net.ApiClient(
            configuration=cfg,
            header_name="authorization",
            header_value=f"Bearer {self.__api_config.app_key}",
        )


class ActivityApi:
    """Wrapper for Activity Control and State classes.

    The purpose of this class is to have a single ActivityApi, just like Payment/Demand/Network,
    without "estetic" split to Control/State.

    Q: Why?
    A: Because we want to keep internal interface as unified as possible, at least for now -
       - we might want to change this in the future.
    """

    def __init__(self, ya_activity_api: ya_activity.ApiClient):
        self.__control_api = ya_activity.RequestorControlApi(ya_activity_api)
        self.__state_api = ya_activity.RequestorStateApi(ya_activity_api)

    def __getattr__(self, attr_name: str) -> Any:
        try:
            return getattr(self.__control_api, attr_name)
        except AttributeError:
            return getattr(self.__state_api, attr_name)


@no_type_check
def get_requestor_api(cls: Type["Resource"], node: "GolemNode") -> TRequestorApi:
    """Return RequestorApi for a given cls, using class typing.

    This is very ugly, but should work well and simplifies the Resource inheritance.
    If we ever decide this is too ugly, it shouldn"t be hard to get rid of this.

    NOTE: this references only "internal" typing, so is invisible from the interface POV.
    """
    api_type = get_args(cls.__orig_bases__[0])[0]
    if api_type is ya_payment.RequestorApi:
        api = ya_payment.RequestorApi(node._ya_payment_api)
        return api
    elif api_type is ya_market.RequestorApi:
        return ya_market.RequestorApi(node._ya_market_api)
    elif api_type is ActivityApi:
        return ActivityApi(node._ya_activity_api)
    elif api_type is ya_net.RequestorApi:
        return ya_net.RequestorApi(node._ya_net_api)
    raise TypeError("This should never happen")



from typing import Any, AsyncIterator, Callable


class Chain:
    """Wrapper class for mid-level components that utilize the pipes and filters pattern.

    Sample usage::

        async def source() -> AsyncIterator[int]:
            yield 1
            yield 2

        async def int_2_str(numbers: AsyncIterator[int]) -> AsyncIterator[str]:
            async for number in numbers:
                yield str(number)

        async for x in Chain(source, int_2_str):
            # x == "1"
            # x == "2"

    A more Golem-specific usage::

        from golem import GolemNode, RepositoryVmPayload
        from golem.mid import (
            Buffer, Chain, Map,
            default_negotiate, default_create_agreement, default_create_activity
        )

        BUDGET = 1
        PAYLOAD = RepositoryVmPayload("9a3b5d67b0b27746283cb5f287c13eab1beaa12d92a9f536b747c7ae")

        async def main():
            async with GolemNode() as golem:
                allocation = await golem.create_allocation(BUDGET)
                demand = await golem.create_demand(PAYLOAD, allocations=[allocation])

                async for activity in Chain(
                    demand.initial_proposals(),
                    Map(default_negotiate),
                    Map(default_create_agreement),
                    Map(default_create_activity),
                    Buffer(),
                ):
                    print(activity)

    """

    def __init__(
        self,
        chain_start: AsyncIterator[Any],
        *pipes: Callable[[AsyncIterator[Any]], AsyncIterator[Any]],
    ):
        aiter = chain_start

        for pipe in pipes:
            aiter = pipe(aiter)

        self._aiter = aiter

    def __aiter__(self) -> "Chain":
        return self

    async def __anext__(self) -> Any:
        return await self._aiter.__anext__()



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Literal, MutableSequence, Union

from golem.payload.mixins import PropsConsSerializerMixin

PropertyName = str
PropertyValue = Any


class ConstraintException(Exception):
    pass


ConstraintOperator = Literal["=", "<=", ">=", "<", ">"]
ConstraintGroupOperator = Literal["&", "|", "!"]


@dataclass
class PayloadSyntaxElement(PropsConsSerializerMixin, ABC):
    def __post_init__(self) -> None:
        self._validate()

    def serialize(self) -> str:
        self._validate()

        return self._serialize()

    @abstractmethod
    def _serialize(self) -> str:
        ...

    def _validate(self) -> None:
        ...


@dataclass
class Constraint(PayloadSyntaxElement):
    property_name: PropertyName
    operator: ConstraintOperator
    value: PropertyValue

    def _serialize(self) -> str:
        serialized_value = self._serialize_value(self.value)

        if not self.value:
            return ""

        if isinstance(self.value, (list, tuple)):
            return ConstraintGroup(
                [Constraint(self.property_name, self.operator, v) for v in serialized_value]
            ).serialize()

        return f"({self.property_name}{self.operator}{serialized_value})"


@dataclass
class ConstraintGroup(PayloadSyntaxElement):
    items: MutableSequence[Union["ConstraintGroup", Constraint]] = field(default_factory=list)
    operator: ConstraintGroupOperator = "&"

    def _validate(self) -> None:
        if self.operator == "!" and 2 <= len(self.items):
            raise ConstraintException("ConstraintGroup with `!` operator can contain only 1 item!")

    def _serialize(self) -> str:
        serialized = [item.serialize() for item in self.items]
        items = "\n\t".join(s for s in serialized if s)

        return f"({self.operator}{items})"


@dataclass
class Constraints(ConstraintGroup):
    pass



from random import random
from typing import Sequence, Tuple, Union

from golem.managers.base import ProposalScorer, ProposalScoringResult
from golem.resources import ProposalData

PropertyValueNumeric = Union[int, float]
BoundaryValues = Tuple[Tuple[float, PropertyValueNumeric], Tuple[float, PropertyValueNumeric]]


class RandomScore(ProposalScorer):
    def __call__(self, proposals_data: Sequence[ProposalData]) -> ProposalScoringResult:
        return [random() for _ in range(len(proposals_data))]



from golem.managers.work.concurrent import ConcurrentWorkManager
from golem.managers.work.mixins import WorkManagerPluginsMixin
from golem.managers.work.plugins import redundancy_cancel_others_on_first_done, retry, work_plugin
from golem.managers.work.sequential import SequentialWorkManager

__all__ = (
    "ConcurrentWorkManager",
    "WorkManagerPluginsMixin",
    "SequentialWorkManager",
    "work_plugin",
    "redundancy_cancel_others_on_first_done",
    "retry",
)



from abc import ABC
from typing import TYPE_CHECKING

from golem.event_bus import Event

if TYPE_CHECKING:
    from golem.node import GolemNode


class GolemNodeEvent(Event, ABC):
    """Base class for all events related to a particular :any:`GolemNode` only."""

    def __init__(self, node: "GolemNode"):
        self._node = node

    @property
    def node(self) -> "GolemNode":
        return self._node

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}({self.node.app_key},"
            f" app_session_id: {self.node.app_session_id})"
        )


class SessionStarted(GolemNodeEvent):
    """Emitted when a :any:`GolemNode` starts operating."""


class ShutdownStarted(GolemNodeEvent):
    """:any:`GolemNode` is closing."""


class ShutdownFinished(GolemNodeEvent):
    """:any:`GolemNode` closed successfully."""



from golem.pipeline.buffer import Buffer
from golem.pipeline.chain import Chain
from golem.pipeline.default_payment_handler import DefaultPaymentHandler
from golem.pipeline.exceptions import InputStreamExhausted
from golem.pipeline.limit import Limit
from golem.pipeline.map import Map
from golem.pipeline.sort import Sort
from golem.pipeline.zip import Zip

__all__ = (
    "Chain",
    "Limit",
    "Map",
    "Zip",
    "Buffer",
    "Sort",
    "InputStreamExhausted",
    "DefaultPaymentHandler",
)



from golem.exceptions import GolemException


class PayloadException(GolemException):
    pass


class ConstraintException(PayloadException):
    pass


class InvalidProperties(PayloadException):
    """`properties` given to `Payload.from_properties(cls, properties)` are invalid."""



from typing import Callable, List, Optional, Sequence, Tuple, Union

from golem.managers.base import ProposalScorer, ProposalScoringResult
from golem.resources import ProposalData

PropertyValueNumeric = Union[int, float]
BoundaryValues = Tuple[Tuple[float, PropertyValueNumeric], Tuple[float, PropertyValueNumeric]]


class MapScore(ProposalScorer):
    def __init__(
        self,
        callback: Callable[[ProposalData], Optional[float]],
        normalize: bool = False,
        normalize_flip: bool = False,
    ) -> None:
        self._callback = callback
        self._normalize = normalize
        self._normalize_flip = normalize_flip

    def __call__(self, proposals_data: Sequence[ProposalData]) -> ProposalScoringResult:
        result = [self._callback(proposal_data) for proposal_data in proposals_data]

        if not self._normalize or result is None:
            return result

        filtered = list(filter(None, result))
        if not filtered:
            return result

        result_max = max(filtered)
        result_min = min(filtered)
        result_div = result_max - result_min

        if result_div == 0:
            return result

        normalized_result: List[Optional[float]] = []
        for v in result:
            if v is not None:
                normalized_result.append((v - result_min) / result_div)
            else:
                normalized_result.append(v)

        if not self._normalize_flip:
            return normalized_result

        flipped_result: List[Optional[float]] = []
        for v in normalized_result:
            if v is not None:
                flipped_result.append(1 - v)
            else:
                flipped_result.append(v)
        return flipped_result



from golem.resources.pooling_batch.events import BatchFinished, NewPoolingBatch
from golem.resources.pooling_batch.exceptions import (
    BatchError,
    BatchTimeoutError,
    CommandCancelled,
    CommandFailed,
    PoolingBatchException,
)
from golem.resources.pooling_batch.pooling_batch import PoolingBatch

__all__ = (
    "PoolingBatch",
    "NewPoolingBatch",
    "BatchFinished",
    "PoolingBatchException",
    "BatchError",
    "CommandFailed",
    "CommandCancelled",
    "BatchTimeoutError",
)



from typing import TYPE_CHECKING

from golem.resources.events import NewResource, ResourceClosed, ResourceDataChanged

if TYPE_CHECKING:
    from golem.resources.activity.activity import Activity  # noqa


class NewActivity(NewResource["Activity"]):
    pass


class ActivityDataChanged(ResourceDataChanged["Activity"]):
    pass


class ActivityClosed(ResourceClosed["Activity"]):
    pass



from golem.resources.activity import (
    Activity,
    ActivityClosed,
    ActivityDataChanged,
    Command,
    Deploy,
    DownloadFile,
    NewActivity,
    Run,
    Script,
    SendFile,
    Start,
    default_prepare_activity,
)
from golem.resources.agreement import (
    Agreement,
    AgreementClosed,
    AgreementDataChanged,
    NewAgreement,
    default_create_activity,
)
from golem.resources.allocation import (
    Allocation,
    AllocationClosed,
    AllocationDataChanged,
    AllocationException,
    NewAllocation,
    NoMatchingAccount,
)
from golem.resources.base import Resource, TResource
from golem.resources.debit_note import (
    DebitNote,
    DebitNoteClosed,
    DebitNoteDataChanged,
    DebitNoteEventCollector,
    NewDebitNote,
)
from golem.resources.demand import (
    Demand,
    DemandBuilder,
    DemandClosed,
    DemandData,
    DemandDataChanged,
    NewDemand,
)
from golem.resources.events import (
    NewResource,
    ResourceClosed,
    ResourceDataChanged,
    ResourceEvent,
    TResourceEvent,
)
from golem.resources.exceptions import ResourceException, ResourceNotFound
from golem.resources.invoice import (
    Invoice,
    InvoiceClosed,
    InvoiceDataChanged,
    InvoiceEventCollector,
    NewInvoice,
)
from golem.resources.network import (
    DeployArgsType,
    Network,
    NetworkClosed,
    NetworkDataChanged,
    NetworkException,
    NetworkFull,
    NewNetwork,
)
from golem.resources.pooling_batch import (
    BatchError,
    BatchFinished,
    BatchTimeoutError,
    CommandCancelled,
    CommandFailed,
    NewPoolingBatch,
    PoolingBatch,
    PoolingBatchException,
)
from golem.resources.proposal import (
    NewProposal,
    Proposal,
    ProposalClosed,
    ProposalData,
    ProposalDataChanged,
    ProposalId,
    default_create_agreement,
    default_negotiate,
)
from golem.resources.utils.payment import LinearCoeffs

__all__ = (
    "Activity",
    "NewActivity",
    "ActivityDataChanged",
    "ActivityClosed",
    "Command",
    "Script",
    "Deploy",
    "Start",
    "Run",
    "SendFile",
    "DownloadFile",
    "default_prepare_activity",
    "Agreement",
    "NewAgreement",
    "AgreementDataChanged",
    "AgreementClosed",
    "default_create_activity",
    "Allocation",
    "AllocationException",
    "NoMatchingAccount",
    "NewAllocation",
    "AllocationDataChanged",
    "AllocationClosed",
    "DebitNote",
    "NewDebitNote",
    "DebitNoteEventCollector",
    "DebitNoteDataChanged",
    "DebitNoteClosed",
    "Demand",
    "DemandBuilder",
    "DemandData",
    "NewDemand",
    "DemandDataChanged",
    "DemandClosed",
    "Invoice",
    "InvoiceEventCollector",
    "NewInvoice",
    "InvoiceDataChanged",
    "InvoiceClosed",
    "Network",
    "DeployArgsType",
    "NetworkException",
    "NetworkFull",
    "NewNetwork",
    "NetworkDataChanged",
    "NetworkClosed",
    "PoolingBatch",
    "NewPoolingBatch",
    "BatchFinished",
    "PoolingBatchException",
    "BatchError",
    "CommandFailed",
    "CommandCancelled",
    "BatchTimeoutError",
    "Proposal",
    "ProposalId",
    "ProposalData",
    "NewProposal",
    "ProposalDataChanged",
    "ProposalClosed",
    "default_negotiate",
    "default_create_agreement",
    "LinearCoeffs",
    "TResourceEvent",
    "TResource",
    "Resource",
    "ResourceEvent",
    "NewResource",
    "ResourceDataChanged",
    "ResourceClosed",
    "ResourceException",
    "ResourceNotFound",
)



import asyncio
from datetime import timedelta
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Union

from ya_activity import models

from golem.resources.base import _NULL, Resource
from golem.resources.pooling_batch.events import BatchFinished, NewPoolingBatch
from golem.resources.pooling_batch.exceptions import (
    BatchError,
    BatchTimeoutError,
    CommandCancelled,
    CommandFailed,
)
from golem.utils.low import ActivityApi, YagnaEventCollector

if TYPE_CHECKING:
    from golem.node import GolemNode
    from golem.resources.activity.activity import Activity  # noqa


class PoolingBatch(
    Resource[ActivityApi, _NULL, "Activity", _NULL, models.ExeScriptCommandResult],
    YagnaEventCollector,
):
    """A single batch of commands.

    Usage::

        batch = activity.execute_commands(Deploy(), Start())
        await batch.wait()
        for event in batch.events:
            print(event.stdout)
    """

    _event_collecting_task: Optional[asyncio.Task] = None

    def __init__(self, node: "GolemNode", id_: str):
        super().__init__(node, id_)
        asyncio.create_task(node.event_bus.emit(NewPoolingBatch(self)))

        self.finished_event = asyncio.Event()
        self._futures: Optional[List[asyncio.Future[models.ExeScriptCommandResult]]] = None

    @property
    def done(self) -> bool:
        """True if this batch is already finished."""
        return self.finished_event.is_set()

    @property
    def success(self) -> bool:
        """True if this batch finished without errors. Raises `AttributeError` if batch is not \
        :any:`done`."""
        if not self.done:
            raise AttributeError("Success can be determined only for finished batches")
        if not self.events:
            #   We got no events but we're done -> only possibility is event collection failure
            #   (In the future maybe also cancelled batch?)
            return False

        return self.events[-1].result == "Ok"

    async def wait(
        self,
        timeout: Optional[Union[timedelta, float]] = None,
        ignore_errors: bool = False,
    ) -> List[models.ExeScriptCommandResult]:
        """Wait until the batch is :any:`done`.

        :param timeout: If not None, :any:`BatchTimeoutError` will be raised if the batch runs
            longer. This exception doesn't stop the execution of the batch.
        :param ignore_errors: When True, :func:`wait` doesn't care if the end of the batch is a
            :any:`success` or not. If False, :any:`BatchError` will be raised for failed batches.
        """
        timeout_seconds: Optional[float]
        if timeout is None:
            timeout_seconds = None
        elif isinstance(timeout, timedelta):
            timeout_seconds = timeout.total_seconds()
        else:
            timeout_seconds = timeout

        try:
            await asyncio.wait_for(self.finished_event.wait(), timeout_seconds)
            if not ignore_errors and not self.success:
                raise BatchError(self)
            return self.events
        except asyncio.TimeoutError:
            assert timeout_seconds is not None  # mypy
            raise BatchTimeoutError(self, timeout_seconds)

    @property
    def events(self) -> List[models.ExeScriptCommandResult]:
        """Returns a list of results for this batch.

        Nth element in the returned list corresponds to the nth command in a batch.

        When the batch is :any:`done`:

        * If the batch is a :any:`success`, there will be a single event for every command.
        * If not, there will be a single event for every successful command followed by
          an event for the command that failed.

        If the batch is not :any:`done`, events for already finished commands will be returned.
        """
        return super().events

    ###########################
    #   Event collector methods
    async def _collect_yagna_events(self) -> None:
        try:
            await super()._collect_yagna_events()
        except Exception:
            #   This happens when activity is destroyed when we're waiting for batch results
            #   (I'm not sure if always - for sure when provider destroys activity because
            #   agreement timed out). Maybe some other scenarios are also possible.
            await self._set_finished()

    def _collect_events_kwargs(self) -> Dict:
        return {"timeout": 5, "_request_timeout": 5.5}

    def _collect_events_args(self) -> List:
        return [self.parent.id, self.id]

    @property
    def _collect_events_func(self) -> Callable:
        return self.api.get_exec_batch_results

    async def _process_event(self, event: models.ExeScriptCommandResult) -> None:
        if event.index < len(self.events):
            #   Repeated event
            return

        self.add_event(event)

        if self._futures is not None:
            if event.result == "Error":
                self._futures[event.index].set_exception(CommandFailed(self))
                for fut in self._futures[event.index + 1 :]:
                    fut.set_exception(CommandCancelled(self))
            else:
                self._futures[event.index].set_result(event)

        if event.is_batch_finished:
            await self._set_finished()

    async def _set_finished(self) -> None:
        await self.node.event_bus.emit(BatchFinished(self))
        self.finished_event.set()
        self.parent.running_batch_counter -= 1
        self.stop_collecting_events()



from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Optional

from golem.payload.properties import Properties

AgreementId = str


@dataclass
class AgreementData:
    agreement_id: AgreementId
    provider_id: Optional[str]
    approved_date: Optional[datetime]
    properties: Properties
    agreement_duration: Optional[timedelta]



import asyncio
from datetime import datetime, timedelta, timezone
from decimal import Decimal
from typing import TYPE_CHECKING, Optional

from ya_payment import RequestorApi, models

from golem.payload import GenericPayload, Payload
from golem.resources.allocation.events import NewAllocation
from golem.resources.allocation.exceptions import NoMatchingAccount
from golem.resources.base import _NULL, Resource, api_call_wrapper
from golem.resources.events import ResourceClosed

if TYPE_CHECKING:
    from golem.node import GolemNode


class Allocation(Resource[RequestorApi, models.Allocation, _NULL, _NULL, _NULL]):
    """A single allocation on the Golem Network.

    Created with one of the :class:`Allocation`-returning methods of the :any:`GolemNode`.
    """

    def __init__(self, node: "GolemNode", id_: str, data: Optional[models.Allocation] = None):
        super().__init__(node, id_, data)
        asyncio.create_task(node.event_bus.emit(NewAllocation(self)))

    @api_call_wrapper(ignore_status_codes=(404, 410))
    async def release(self) -> None:
        """Release the allocation.

        Remaining amount will be available again. This is the final operation -
        released allocation is not available anymore.
        """
        await self.api.release_allocation(self.id)
        await self.node.event_bus.emit(ResourceClosed(self))

    @classmethod
    async def create_any_account(
        cls,
        node: "GolemNode",
        amount: Decimal,
        network: str,
        driver: str,
    ) -> "Allocation":
        for account in await cls._get_api(node).get_requestor_accounts():
            if (
                account.driver.lower() == driver.lower()
                and account.network.lower() == network.lower()
            ):
                break
        else:
            raise NoMatchingAccount(network, driver)

        return await cls.create_with_account(node, account, amount)

    @classmethod
    async def create_with_account(
        cls,
        node: "GolemNode",
        account: models.Account,
        amount: Decimal,
    ) -> "Allocation":
        timestamp = datetime.now(timezone.utc)
        timeout = timestamp + timedelta(days=365 * 10)

        data = models.Allocation(
            address=account.address,
            payment_platform=account.platform,
            total_amount=str(amount),
            timestamp=timestamp,
            timeout=timeout,
            #   This will probably be removed one day (consent-related thing)
            make_deposit=False,
            #   We must set this here because of the ya_client interface
            allocation_id="",
            spent_amount="",
            remaining_amount="",
        )

        return await cls.create(node, data)

    @classmethod
    async def create(cls, node: "GolemNode", data: models.Allocation) -> "Allocation":
        api = cls._get_api(node)
        created = await api.create_allocation(data)
        return cls(node, created.allocation_id, created)

    @api_call_wrapper()
    async def get_demand_spec(self) -> Payload:
        data = await self.api.get_demand_decorations([self.id])
        return GenericPayload.from_raw_api_data(data.properties, data.constraints)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from golem.payload import Constraints, Properties


@dataclass
class DemandData:
    properties: Properties
    constraints: Constraints
    demand_id: Optional[str]
    requestor_id: Optional[str]
    timestamp: datetime



import asyncio
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, AsyncIterator, Optional, Union, cast

from ya_market import RequestorApi
from ya_market import models as models

from golem.payload import Constraints, NodeInfo, PayloadSyntaxParser, Properties
from golem.resources.agreement import Agreement
from golem.resources.base import Resource, api_call_wrapper
from golem.resources.proposal.data import ProposalData
from golem.resources.proposal.events import NewProposal
from golem.resources.proposal.exceptions import ProposalRejected

if TYPE_CHECKING:
    from golem.node import GolemNode
    from golem.resources.demand import Demand

DEFAULT_TTL = timedelta(hours=1)


class Proposal(
    Resource[
        RequestorApi,
        models.Proposal,
        Union["Demand", "Proposal"],
        Union["Proposal", Agreement],
        Union[models.ProposalEvent, models.ProposalRejectedEvent],
    ]
):
    """A single proposal on the Golem Network.

    Either a initial proposal matched to a demand, or a counter-proposal sent
    either by us or by the provider.

    Sample usage::

        initial_proposal = await demand.initial_proposals().__anext__()
        our_counter_proposal = await initial_proposal.respond()
        async for their_counter_proposal in our_counter_proposal.responses():
            agreement = their_counter_proposal.create_agreement()
            break
        else:
            print("Our counter-proposal was rejected :(")
    """

    _demand: Optional["Demand"] = None
    _proposal_data: Optional[ProposalData] = None
    _provider_node_name: Optional[str] = None

    def __init__(self, node: "GolemNode", id_: str, data: Optional[models.Proposal] = None):
        super().__init__(node, id_, data)
        asyncio.create_task(node.event_bus.emit(NewProposal(self)))

    ##############################
    #   State-related properties
    @property
    def initial(self) -> bool:
        """True for proposals matched directly to the demand."""
        return self.parent == self.demand

    @property
    def draft(self) -> bool:
        """True for proposals that are responses to other proposals."""
        assert self.data is not None
        return self.data.state == "Draft"

    @property
    def rejected(self) -> bool:
        """True for rejected proposals. They will have no more :func:`responses`."""
        assert self.data is not None
        return self.data.state == "Rejected"

    ###########################
    #   Tree-related methods
    @property
    def demand(self) -> "Demand":
        """Initial :class:`Demand` of this proposal."""
        # We can either have no parent (this is possible when this Proposal was created from id),
        # and then _demand is always set, or a Proposal-parent or a Demand-parent.

        # FIXME: remove local import
        from golem.resources import Demand

        if self._demand is not None:
            return self._demand
        else:
            if isinstance(self.parent, Demand):
                return self.parent
            else:
                return self.parent.demand  # TODO recursion

    @demand.setter
    def demand(self, demand: "Demand") -> None:
        assert self._demand is None
        assert (
            self._parent is None
        )  # Sanity check (there's no scenario where we have a parent and demand is set)
        self._demand = demand

    def add_event(self, event: Union[models.ProposalEvent, models.ProposalRejectedEvent]) -> None:
        super().add_event(event)
        if isinstance(event, models.ProposalRejectedEvent):
            self.set_no_more_children(ProposalRejected(event.proposal_id, event.reason.message))

    async def responses(self) -> AsyncIterator["Proposal"]:
        """Yield responses to this proposal.

        Stops when the proposal is rejected.
        """
        async for child in self.child_aiter():
            if isinstance(child, Proposal):
                yield child

    ############################
    #   Negotiations
    @api_call_wrapper()
    async def create_agreement(
        self, autoclose: bool = True, timeout: timedelta = timedelta(seconds=60)
    ) -> "Agreement":
        """Promote this proposal to an agreement.

        :param autoclose: Terminate the agreement when the :any:`GolemNode` closes.
        :param timeout: TODO - this is used as `AgreementValidTo`, but what is it exactly?
        """
        proposal = models.AgreementProposal(
            proposal_id=self.id,
            # TODO: what is AgreementValidTo?
            valid_to=datetime.now(timezone.utc) + timeout,  # type: ignore
        )
        agreement_id = await self.api.create_agreement(proposal)
        agreement = Agreement(self.node, agreement_id)
        self.add_child(agreement)
        if autoclose:
            self.node.add_autoclose_resource(agreement)

        return agreement

    @api_call_wrapper()
    async def reject(self, reason: str = "") -> None:
        """Reject the proposal - inform the provider that we won't send any more counter-proposals.

        :param reason: An optional information for the provider describing rejection reasons.

        Invalid on our responses.
        """
        await self.api.reject_proposal_offer(
            self.demand.id, self.id, request_body={"message": reason}, _request_timeout=5
        )

    @api_call_wrapper()
    async def respond(
        self, properties: Optional[Properties] = None, constraints: Optional[Constraints] = None
    ) -> "Proposal":
        """Respond to a proposal with a counter-proposal.

        Invalid on our responses.

        Related issues:
        https://github.com/golemfactory/golem-core-python/issues/17
        https://github.com/golemfactory/golem-core-python/issues/18
        """
        if properties is None and constraints is None:
            data = await self._response_data()
        elif properties is not None and constraints is not None:
            data = models.DemandOfferBase(
                properties=properties.serialize(), constraints=constraints.serialize()
            )
        else:
            raise ValueError("Both `properties` and `constraints` arguments must be provided!")

        new_proposal_id = await self.api.counter_proposal_demand(
            self.demand.id, self.id, data, _request_timeout=5
        )

        new_proposal = type(self)(self.node, new_proposal_id)
        self.add_child(new_proposal)

        return new_proposal

    async def _response_data(self) -> models.DemandOfferBase:
        demand_data = await self.demand.get_data()
        data = models.DemandOfferBase(
            properties=demand_data.properties, constraints=demand_data.constraints
        )
        return data

    ##########################
    #   Other
    async def _get_data(self) -> models.Proposal:
        assert self.demand is not None
        data: models.Proposal = await self.api.get_proposal_offer(self.demand.id, self.id)
        if data.state == "Rejected":
            self.set_no_more_children(
                ProposalRejected(data.proposal_id, "Proposal found at `Rejected` state")
            )
        return data

    @classmethod
    def from_proposal_event(cls, node: "GolemNode", event: models.ProposalEvent) -> "Proposal":
        data = event.proposal
        assert data.proposal_id is not None  # mypy
        proposal = Proposal(node, data.proposal_id, data)
        proposal.add_event(event)
        return proposal

    async def get_proposal_data(self) -> ProposalData:
        if not self._proposal_data:
            data = await self.get_data()
            constraints = PayloadSyntaxParser.get_instance().parse_constraints(data.constraints)
            self._proposal_data = ProposalData(
                properties=Properties(data.properties),
                constraints=constraints,
                proposal_id=data.proposal_id,
                issuer_id=data.issuer_id,
                state=data.state,
                timestamp=cast(datetime, data.timestamp),
                prev_proposal_id=data.prev_proposal_id,
            )

        return self._proposal_data

    async def get_provider_id(self):
        """Get the node id of the provider which issued this Proposal."""
        proposal_data = await self.get_data()
        return proposal_data.issuer_id

    async def get_provider_name(self):
        """Get the node name of the provider which issued this Proposal."""
        if not self._provider_node_name:
            proposal_data = await self.get_data()
            node_info = NodeInfo.from_properties(proposal_data.properties)
            self._provider_node_name = node_info.name
        return self._provider_node_name

    async def get_expiration_date(self) -> datetime:
        """Return expiration date to auto unsubscribe.

        Note: As Proposal can have different expiration date than its Demand, it would be unusable
        after demand expiration anyway, hence earliest from both dates is returned.
        """
        await self.get_data()

        demand_expiration_date = await self.demand.get_expiration_date()
        proposal_expiration_date = cast(datetime, self.data.timestamp) + DEFAULT_TTL

        return min(proposal_expiration_date, demand_expiration_date)



from golem.utils.asyncio.buffer import (
    BackgroundFillBuffer,
    Buffer,
    ComposableBuffer,
    ExpirableBuffer,
    SimpleBuffer,
)
from golem.utils.asyncio.queue import ErrorReportingQueue
from golem.utils.asyncio.semaphore import SingleUseSemaphore
from golem.utils.asyncio.tasks import (
    create_task_with_logging,
    ensure_cancelled,
    ensure_cancelled_many,
)
from golem.utils.asyncio.waiter import Waiter

__all__ = (
    "BackgroundFillBuffer",
    "Buffer",
    "ComposableBuffer",
    "ExpirableBuffer",
    "SimpleBuffer",
    "ErrorReportingQueue",
    "SingleUseSemaphore",
    "ensure_cancelled",
    "ensure_cancelled_many",
    "create_task_with_logging",
    "Waiter",
)



"""Storage models."""

import abc
import asyncio
import io
import os
import pathlib
from os import PathLike
from typing import AsyncIterator, NamedTuple, Optional, Union

import aiohttp

_BUF_SIZE = 40960
DOWNLOAD_BYTES_LIMIT_DEFAULT = 1 * 1024 * 1024
AsyncReader = Union[asyncio.streams.StreamReader, aiohttp.streams.StreamReader]


class Content(NamedTuple):
    length: int
    stream: AsyncIterator[bytes]

    @classmethod
    def from_reader(cls, length: int, s: AsyncReader) -> "Content":
        async def stream() -> AsyncIterator[bytes]:
            while not s.at_eof():
                buf = await s.read(_BUF_SIZE)
                yield buf

        return Content(length, stream())


class Source(abc.ABC):
    @property
    @abc.abstractmethod
    def download_url(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    async def content_length(self) -> int:
        raise NotImplementedError


class Destination(abc.ABC):
    @property
    @abc.abstractmethod
    def upload_url(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    async def download_stream(self) -> Content:
        raise NotImplementedError

    async def download_bytes(self, limit: int = DOWNLOAD_BYTES_LIMIT_DEFAULT) -> bytes:
        output = b""
        content = await self.download_stream()

        async for chunk in content.stream:
            limit_remaining = limit - len(output)
            if limit_remaining > len(chunk):
                output += chunk
            else:
                output += chunk[:limit_remaining]
                break

        return output

    async def download_file(self, destination_file: PathLike) -> None:
        content = await self.download_stream()
        with open(destination_file, "wb") as f:
            async for chunk in content.stream:
                f.write(chunk)


class InputStorageProvider(abc.ABC):
    @abc.abstractmethod
    async def upload_stream(self, length: int, stream: AsyncIterator[bytes]) -> Source:
        raise NotImplementedError

    async def upload_bytes(self, data: bytes) -> Source:
        async def _inner() -> AsyncIterator[bytes]:
            yield data

        return await self.upload_stream(len(data), _inner())

    async def upload_file(self, path: os.PathLike) -> Source:
        fp = pathlib.Path(path)
        file_size = fp.stat().st_size

        async def read_file() -> AsyncIterator[bytes]:
            with io.open(path, "rb") as f:
                while True:
                    b: bytes = f.read(_BUF_SIZE)
                    if not b:
                        break
                    yield b

        return await self.upload_stream(file_size, read_file())

    async def release_source(self, source: Source) -> None:
        """Release a source returned by `upload_file` or `upload_bytes`.

        The default implementation is to do nothing.
        """


class OutputStorageProvider(abc.ABC):
    @abc.abstractmethod
    async def new_destination(self, destination_file: Optional[PathLike] = None) -> Destination:
        """Create slot for receiving file.

        Parameters
        ----------
        destination_file:
            Optional hint where received data should be placed.

        """
        raise NotImplementedError


class StorageProvider(InputStorageProvider, OutputStorageProvider, abc.ABC):
    pass


class ComposedStorageProvider(StorageProvider):
    def __init__(self, input_storage: InputStorageProvider, output_storage: OutputStorageProvider):
        self._input = input_storage
        self._output = output_storage

    async def upload_stream(self, length: int, stream: AsyncIterator[bytes]) -> Source:
        return await self._input.upload_stream(length, stream)

    async def upload_file(self, path: os.PathLike) -> Source:
        return await self._input.upload_file(path)

    async def new_destination(self, destination_file: Optional[PathLike] = None) -> Destination:
        return await self._output.new_destination(destination_file)



from abc import ABC, abstractmethod
from typing import Awaitable, Callable, Generic, Optional, Type, TypeVar

from golem.exceptions import GolemException

TCallbackHandler = TypeVar("TCallbackHandler")

TEvent = TypeVar("TEvent", bound="Event")


class Event(ABC):
    """Base class for all events."""


class EventBusError(GolemException):
    pass


class EventBus(ABC, Generic[TCallbackHandler]):
    @abstractmethod
    async def start(self) -> None:
        ...

    @abstractmethod
    async def stop(self) -> None:
        ...

    @abstractmethod
    def is_started(self) -> bool:
        ...

    @abstractmethod
    async def on(
        self,
        event_type: Type[TEvent],
        callback: Callable[[TEvent], Awaitable[None]],
        filter_func: Optional[Callable[[TEvent], bool]] = None,
    ) -> TCallbackHandler:
        ...

    @abstractmethod
    async def on_once(
        self,
        event_type: Type[TEvent],
        callback: Callable[[TEvent], Awaitable[None]],
        filter_func: Optional[Callable[[TEvent], bool]] = None,
    ) -> TCallbackHandler:
        ...

    @abstractmethod
    async def off(self, callback_handler: TCallbackHandler) -> None:
        ...

    @abstractmethod
    async def emit(self, event: TEvent) -> None:
        ...



from datetime import datetime
from typing import List, Optional, Sequence, Tuple, cast

from golem.managers.base import ScorerWithOptionalWeight
from golem.payload import PayloadSyntaxParser, Properties
from golem.resources import Proposal, ProposalData
from golem.utils.asyncio.tasks import resolve_maybe_awaitable
from golem.utils.logging import trace_span


class ProposalScoringMixin:
    def __init__(
        self,
        proposal_scorers: Optional[Sequence[ScorerWithOptionalWeight]] = None,
        *args,
        **kwargs,
    ) -> None:
        self._demand_offer_parser = PayloadSyntaxParser.get_instance()
        self._proposal_scorers: List[ScorerWithOptionalWeight] = (
            list(proposal_scorers) if proposal_scorers is not None else []
        )

        super().__init__(*args, **kwargs)

    async def do_scoring(self, proposals: Sequence[Proposal]) -> List[Tuple[float, Proposal]]:
        proposals_data = await self._get_proposals_data_from_proposals(proposals)
        proposal_scores = await self._run_scorers(proposals_data)

        scored_proposals = self._calculate_proposal_score(proposals, proposal_scores)
        scored_proposals.sort(key=lambda x: x[0], reverse=True)

        return scored_proposals

    @trace_span()
    async def _run_scorers(
        self, proposals_data: Sequence[ProposalData]
    ) -> Sequence[Tuple[float, Sequence[float]]]:
        proposal_scores: List[Tuple[float, Sequence[float]]] = []

        for scorer in self._proposal_scorers:
            if isinstance(scorer, (list, tuple)):
                weight, scorer = scorer
            else:
                weight = 1

            scorer_scores = await resolve_maybe_awaitable(scorer(proposals_data))

            proposal_scores.append((weight, scorer_scores))  # type: ignore[arg-type]

        return proposal_scores

    def _calculate_proposal_score(
        self,
        proposals: Sequence[Proposal],
        scorer_scores: Sequence[Tuple[float, Sequence[float]]],
    ) -> List[Tuple[float, Proposal]]:
        # FIXME: can this be refactored?
        return [
            (
                self._calculate_weighted_score(
                    self._transpose_scorer_scores(proposal_index, scorer_scores)
                ),
                proposal,
            )
            for proposal_index, proposal in enumerate(proposals)
        ]

    def _calculate_weighted_score(
        self, proposal_weighted_scores: Sequence[Tuple[float, float]]
    ) -> float:
        if not proposal_weighted_scores:
            return 0

        weighted_sum = sum(pws[0] * pws[1] for pws in proposal_weighted_scores)
        weights_sum = sum(pws[0] for pws in proposal_weighted_scores)

        return weighted_sum / weights_sum

    def _transpose_scorer_scores(
        self, proposal_index: int, scorer_scores: Sequence[Tuple[float, Sequence[float]]]
    ) -> Sequence[Tuple[float, float]]:
        # FIXME: can this be refactored?
        return [
            (scorer_weight, scorer_scores[proposal_index])
            for scorer_weight, scorer_scores in scorer_scores
            if scorer_scores[proposal_index] is not None
        ]

    # FIXME: This should be already provided by low level
    async def _get_proposals_data_from_proposals(
        self, proposals: Sequence[Proposal]
    ) -> Sequence[ProposalData]:
        result = []

        for proposal in proposals:
            data = await proposal.get_data()

            constraints = self._demand_offer_parser.parse_constraints(data.constraints)

            result.append(
                ProposalData(
                    properties=Properties(data.properties),
                    constraints=constraints,
                    proposal_id=data.proposal_id,
                    issuer_id=data.issuer_id,
                    state=data.state,
                    timestamp=cast(datetime, data.timestamp),
                    prev_proposal_id=data.prev_proposal_id,
                )
            )

        return result



from typing import TYPE_CHECKING

from golem.resources.events import NewResource, ResourceClosed, ResourceDataChanged

if TYPE_CHECKING:
    from golem.resources.agreement.agreement import Agreement  # noqa


class NewAgreement(NewResource["Agreement"]):
    pass


class AgreementDataChanged(ResourceDataChanged["Agreement"]):
    pass


class AgreementClosed(ResourceClosed["Agreement"]):
    pass



from golem.managers.proposal.plugins.blacklist import BlacklistProviderIdPlugin
from golem.managers.proposal.plugins.buffer import ProposalBuffer
from golem.managers.proposal.plugins.linear_coeffs import LinearCoeffsCost, LinearPerCpuCoeffsCost
from golem.managers.proposal.plugins.negotiating import (
    MidAgreementPaymentsNegotiator,
    NegotiatingPlugin,
    PaymentPlatformNegotiator,
)
from golem.managers.proposal.plugins.reject_costs_exceeds import RejectIfCostsExceeds
from golem.managers.proposal.plugins.scoring import (
    LinearAverageCostPricing,
    LinearPerCpuAverageCostPricing,
    MapScore,
    PropertyValueLerpScore,
    ProposalScoringBuffer,
    ProposalScoringMixin,
    RandomScore,
)

__all__ = (
    "BlacklistProviderIdPlugin",
    "ProposalBuffer",
    "PaymentPlatformNegotiator",
    "MidAgreementPaymentsNegotiator",
    "NegotiatingPlugin",
    "RejectIfCostsExceeds",
    "MapScore",
    "ProposalScoringMixin",
    "LinearAverageCostPricing",
    "LinearPerCpuAverageCostPricing",
    "LinearCoeffsCost",
    "LinearPerCpuCoeffsCost",
    "PropertyValueLerpScore",
    "RandomScore",
    "ProposalScoringBuffer",
)



from typing import TYPE_CHECKING, Callable, Tuple

from golem.resources.event_collectors import InvoiceEvent, PaymentEventCollector
from golem.resources.invoice.invoice import Invoice

if TYPE_CHECKING:
    from golem.resources.agreement import Agreement


class InvoiceEventCollector(PaymentEventCollector):
    @property
    def _collect_events_func(self) -> Callable:
        return Invoice._get_api(self.node).get_invoice_events

    async def _get_event_resources(self, event: InvoiceEvent) -> Tuple[Invoice, "Agreement"]:
        assert event.invoice_id is not None
        invoice = self.node.invoice(event.invoice_id)
        await invoice.get_data()
        agreement = self.node.agreement(invoice.data.agreement_id)
        return invoice, agreement



import contextvars
import inspect
import logging
import sys
from datetime import datetime, timezone
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Sequence, Union

if TYPE_CHECKING:
    from golem.event_bus import Event


if (3, 12) <= sys.version_info:
    FilterReturn = Union[bool, logging.LogRecord]
else:
    FilterReturn = bool


DEFAULT_LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "filters": {
        "add_trace_id": {
            "()": "golem.utils.logging.AddTraceIdFilter",
        },
    },
    "formatters": {
        "default": {
            "format": "[%(asctime)s] [%(levelname)-7s] [%(traceid)s] "
            "[%(name)s:%(lineno)d] %(message)s",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "default",
            "filters": ["add_trace_id"],
        },
    },
    "root": {
        "level": "INFO",
        "handlers": [
            "console",
        ],
    },
    "loggers": {
        "asyncio": {
            "level": "INFO",
        },
        "golem": {
            "level": "INFO",
        },
        "golem.utils": {
            "level": "INFO",
        },
        "golem.managers": {
            "level": "INFO",
        },
        "golem.managers.payment": {
            "level": "INFO",
        },
        "golem.managers.network": {
            "level": "INFO",
        },
        "golem.managers.demand": {
            "level": "INFO",
        },
        "golem.managers.negotiation": {
            "level": "INFO",
        },
        "golem.managers.proposal": {
            "level": "INFO",
        },
        "golem.managers.agreement": {
            "level": "INFO",
        },
        "golem.managers.activity": {
            "level": "INFO",
        },
        "golem.managers.work": {
            "level": "INFO",
        },
    },
}


class _YagnaDatetimeFormatter(logging.Formatter):
    """Custom log Formatter that formats datetime using the same convention yagna uses."""

    LOCAL_TZ = datetime.now(timezone.utc).astimezone().tzinfo

    def formatTime(self, record: logging.LogRecord, datefmt: Optional[str] = None) -> str:
        """Format datetime; example: `2021-06-11T14:55:43.156.123+0200`."""
        dt = datetime.fromtimestamp(record.created, tz=self.LOCAL_TZ)
        millis = f"{(dt.microsecond // 1000):03d}"
        return dt.strftime(f"%Y-%m-%dT%H:%M:%S.{millis}%z")


class DefaultLogger:
    """Dump all events to a file.

    Usage::

        golem = GolemNode()
        await golem.event_bus.on(Event, DefaultLogger().on_event)

    Or::

        DefaultLogger().logger.debug("What's up?")
    """

    def __init__(self, file_name: str = "log.log"):
        """Init DefaultLogger.

        :param file_name: Name of the file where all events will be dumped.
        """
        self._file_name = file_name
        self._logger = self._prepare_logger()

    @property
    def file_name(self) -> str:
        """Name of the file where all events will be dumped."""
        return self._file_name

    @property
    def logger(self) -> logging.Logger:
        """Logger that just dumps everything to file :any:`file_name`."""
        return self._logger

    def _prepare_logger(self) -> logging.Logger:
        logger = logging.getLogger("golem")
        logger.setLevel(logging.DEBUG)

        format_ = "[%(asctime)s %(levelname)s %(name)s] %(message)s"
        formatter = _YagnaDatetimeFormatter(fmt=format_)

        file_handler = logging.FileHandler(filename=self._file_name, mode="w", encoding="utf-8")
        file_handler.setFormatter(formatter)
        file_handler.setLevel(logging.DEBUG)
        logger.addHandler(file_handler)

        return logger

    async def on_event(self, event: "Event") -> None:
        """Handle event produced by :any:`EventBus.on`."""
        self.logger.info(event)


trace_id_var = contextvars.ContextVar("trace_id", default="root")


def get_trace_id_name(obj: Any, postfix: str) -> str:
    return f"{obj.__class__.__name__}-{id(obj)}-{postfix}"


class TraceSpan:
    def __init__(
        self,
        name: Optional[Union[str, Callable[[Any], str]]] = None,
        show_arguments: bool = False,
        show_results: bool = False,
        log_level: int = logging.DEBUG,
    ) -> None:
        self._name = name
        self._show_arguments = show_arguments
        self._show_results = show_results
        self._log_level = log_level

    def __call__(self, func):
        wrapper = self._async_wrapper if inspect.iscoroutinefunction(func) else self._sync_wrapper

        # TODO: partial instead of decorator()
        def decorator(*args, **kwargs):
            return wrapper(func, args, kwargs)

        return wraps(func)(decorator)

    def _get_span_name(self, func: Callable, args: Sequence, kwargs: Dict) -> str:
        if self._name is not None:
            return self._name(args[0]) if callable(self._name) else self._name

        # TODO: check type of func in different cases + contextmanager
        span_name = (
            func.__qualname__.split(">.")[-1] if self._is_instance_method(func) else func.__name__
        )

        if self._show_arguments:
            arguments = ", ".join(
                [
                    *[repr(a) for a in (args[1:] if self._is_instance_method(func) else args)],
                    *["{}={}".format(k, repr(v)) for (k, v) in kwargs.items()],
                ]
            )
            return f"Calling {span_name}({arguments})"

        return f"Calling {span_name}"

    def _get_logger(self, func: Callable, args: Sequence) -> logging.Logger:
        module_name = (
            args[0].__class__.__module__ if self._is_instance_method(func) else func.__module__
        )

        return logging.getLogger(module_name)

    def _is_instance_method(self, func: Callable) -> bool:
        return inspect.isfunction(func) and func.__name__ != func.__qualname__.split(">.")[-1]

    def _sync_wrapper(self, func: Callable, args: Sequence, kwargs: Dict) -> Any:
        span_name = self._get_span_name(func, args, kwargs)
        logger = self._get_logger(func, args)

        logger.log(self._log_level, "%s...", span_name)

        try:
            result = func(*args, **kwargs)
        except Exception as e:
            logger.log(self._log_level, "%s failed with `%s`", span_name, e)
            raise

        if self._show_results:
            logger.log(self._log_level, "%s done with `%s`", span_name, result)
        else:
            logger.log(self._log_level, "%s done", span_name)

        return result

    async def _async_wrapper(self, func: Callable, args: Sequence, kwargs: Dict) -> Any:
        span_name = self._get_span_name(func, args, kwargs)
        logger = self._get_logger(func, args)

        logger.log(self._log_level, "%s...", span_name)

        try:
            result = await func(*args, **kwargs)
        except Exception as e:
            logger.log(self._log_level, "%s failed with `%s`", span_name, e)
            raise

        if self._show_results:
            logger.log(self._log_level, "%s done with `%s`", span_name, result)
        else:
            logger.log(self._log_level, "%s done", span_name)

        return result


trace_span = TraceSpan


class AddTraceIdFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> FilterReturn:
        record.traceid = trace_id_var.get()

        return super().filter(record)



from typing import TYPE_CHECKING

from golem.resources.events import NewResource, ResourceClosed, ResourceDataChanged

if TYPE_CHECKING:
    from golem.resources.proposal.proposal import Proposal  # noqa


class NewProposal(NewResource["Proposal"]):
    pass


class ProposalDataChanged(ResourceDataChanged["Proposal"]):
    pass


class ProposalClosed(ResourceClosed["Proposal"]):
    pass



from typing import TYPE_CHECKING

from golem.resources.events import NewResource, ResourceClosed, ResourceDataChanged

if TYPE_CHECKING:
    from golem.resources.demand.demand import Demand  # noqa


class NewDemand(NewResource["Demand"]):
    pass


class DemandDataChanged(ResourceDataChanged["Demand"]):
    pass


class DemandClosed(ResourceClosed["Demand"]):
    pass



import logging
from typing import Set

from golem.managers import ProposalManagerPlugin
from golem.resources import Proposal
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)


class BlacklistProviderIdPlugin(ProposalManagerPlugin):
    def __init__(self, blacklist: Set[str]) -> None:
        self._blacklist = blacklist

    @trace_span(show_results=True)
    async def get_proposal(self) -> Proposal:
        while True:
            proposal: Proposal = await self._get_proposal()
            proposal_data = await proposal.get_data()
            provider_id = proposal_data.issuer_id

            if provider_id not in self._blacklist:
                return proposal

            if not proposal.initial:
                await proposal.reject("provider_id is on blacklist")

            logger.debug(
                "Provider `%s` from proposal `%s` is on blacklist, picking different proposal...",
                provider_id,
                proposal,
            )



import asyncio
import itertools
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import asynccontextmanager
from datetime import timedelta
from typing import (
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    MutableSequence,
    Optional,
    Sequence,
    TypeVar,
)

from golem.utils.asyncio.semaphore import SingleUseSemaphore
from golem.utils.asyncio.tasks import (
    create_delayed_task_with_logging,
    create_task_with_logging,
    ensure_cancelled,
    ensure_cancelled_many,
    resolve_maybe_awaitable,
)
from golem.utils.logging import get_trace_id_name
from golem.utils.typing import MaybeAwaitable

TItem = TypeVar("TItem")

logger = logging.getLogger(__name__)


class Buffer(ABC, Generic[TItem]):
    """Interface class for object similar to `asyncio.Queue` but with more control over its \
    items."""

    # TODO: Consider hiding condition in favor of public lock related context manager method
    condition: asyncio.Condition

    @abstractmethod
    def size(self) -> int:
        """Return the number of items stored in the buffer."""

    @abstractmethod
    async def wait_for_any_items(self, *, lock=True) -> None:
        """Wait until any items are stored in the buffer.

        If `.set_exception()` was previously called, exception should be raised only if buffer
        is empty.
        """

    @abstractmethod
    async def get(self, *, lock=True) -> TItem:
        """Await, remove and return left-most item stored in the buffer.

        If `.set_exception()` was previously called, the exception will be raised only if the buffer
        is empty.
        """

    @abstractmethod
    async def get_all(self, *, lock=True) -> MutableSequence[TItem]:
        """Remove and immediately return all items stored in the buffer.

        Note that this method should not await any items if the buffer is empty and should instead
        return an empty iterable.

        If `.set_exception()` was previously called, exception should be raised only if buffer
        is empty.
        """

    @abstractmethod
    async def put(self, item: TItem, *, lock=True) -> None:
        """Add item to right-most position to buffer.

        Duplicates are supported.
        """

    @abstractmethod
    async def put_all(self, items: Sequence[TItem], *, lock=True) -> None:
        """Replace all items stored in buffer.

        Duplicates are supported.
        """

    @abstractmethod
    async def remove(self, item: TItem, *, lock=True) -> None:
        """Remove the first occurrence of the item from the buffer or raise `ValueError` \
        if not found."""

    @abstractmethod
    async def set_exception(self, exc: BaseException, *, lock=True) -> None:
        """Set the exception that will be raised on `.get()`/`.get_all()` if there are no more \
         items in the buffer."""

    def reset_exception(self) -> None:
        """Reset (clear) the exception that was previously set by calling `.set_exception()`."""


class ComposableBuffer(Buffer[TItem]):
    """Utility class for composable/stackable buffer implementations to help with calling \
    the underlying buffer."""

    def __init__(self, buffer: Buffer[TItem]):
        self._buffer = buffer

    @asynccontextmanager
    async def _handle_lock(self, lock: bool):
        if lock:
            async with self._buffer.condition:
                yield
        else:
            yield

    def size(self) -> int:
        return self._buffer.size()

    async def wait_for_any_items(self, *, lock=True) -> None:
        await self._buffer.wait_for_any_items(lock=lock)

    async def get(self, *, lock=True) -> TItem:
        return await self._buffer.get(lock=lock)

    async def get_all(self, *, lock=True) -> MutableSequence[TItem]:
        return await self._buffer.get_all(lock=lock)

    async def put(self, item: TItem, *, lock=True) -> None:
        await self._buffer.put(item, lock=lock)

    async def put_all(self, items: Sequence[TItem], *, lock=True) -> None:
        await self._buffer.put_all(items, lock=lock)

    async def remove(self, item: TItem, *, lock=True) -> None:
        await self._buffer.remove(item, lock=lock)

    async def set_exception(self, exc: BaseException, *, lock=True) -> None:
        await self._buffer.set_exception(exc, lock=lock)

    def reset_exception(self) -> None:
        self._buffer.reset_exception()


class SimpleBuffer(Buffer[TItem]):
    """Basic implementation of the Buffer interface."""

    def __init__(self, items: Optional[Sequence[TItem]] = None):
        self._items = list(items) if items else []
        self._error: Optional[BaseException] = None

        self.condition = asyncio.Condition()

    def size(self) -> int:
        return len(self._items)

    @asynccontextmanager
    async def _handle_lock(self, lock: bool):
        if lock:
            async with self.condition:
                yield
        else:
            yield

    async def wait_for_any_items(self, lock=True) -> None:
        async with self._handle_lock(lock):
            await self.condition.wait_for(lambda: bool(self.size() or self._error))

            self._handle_error_if_empty()

    def _handle_error_if_empty(self) -> None:
        if not self.size() and self._error:
            raise self._error

    async def get(self, *, lock=True) -> TItem:
        async with self._handle_lock(lock):
            await self.wait_for_any_items(lock=False)

            return self._items.pop(0)

    async def get_all(self, *, lock=True) -> MutableSequence[TItem]:
        async with self._handle_lock(lock):
            self._handle_error_if_empty()

            items = self._items[:]
            self._items.clear()

            return items

    async def put(self, item: TItem, *, lock=True) -> None:
        async with self._handle_lock(lock):
            self._items.append(item)
            self.condition.notify()

    async def put_all(self, items: Sequence[TItem], *, lock=True) -> None:
        async with self._handle_lock(lock):
            self._items.clear()
            self._items.extend(items[:])

            self.condition.notify(len(items))

    async def remove(self, item: TItem, *, lock=True) -> None:
        async with self._handle_lock(lock):
            self._items.remove(item)

    async def set_exception(self, exc: BaseException, *, lock=True) -> None:
        async with self._handle_lock(lock):
            self._error = exc
            self.condition.notify()

    def reset_exception(self) -> None:
        self._error = None


class ExpirableBuffer(ComposableBuffer[TItem]):
    """Composable `Buffer` that adds option to expire item after some time.

    Items that are already in provided buffer will not expire.
    """

    # TODO: Optimisation options: Use single expiration task that wakes up to expire the earliest
    #  item, then check next earliest item and sleep to it and repeat

    def __init__(
        self,
        buffer: Buffer[TItem],
        get_expiration_func: Callable[[TItem], MaybeAwaitable[Optional[timedelta]]],
        on_expired_func: Optional[Callable[[TItem], MaybeAwaitable[None]]] = None,
    ):
        super().__init__(buffer)

        self._get_expiration_func = get_expiration_func
        self._on_expired_func = on_expired_func

        self._expiration_tasks: Dict[int, List[asyncio.Task]] = defaultdict(list)

    async def _add_expiration_task_for_item(self, item: TItem) -> None:
        expiration = await resolve_maybe_awaitable(self._get_expiration_func(item))

        if expiration is None:
            return

        self._expiration_tasks[id(item)].append(
            create_delayed_task_with_logging(
                expiration,
                self._expire_item_shielded(item),
                trace_id=get_trace_id_name(self, f"item-expire-{item}"),
            )
        )

    async def _remove_expiration_task_for_item(self, item: TItem) -> None:
        item_id = id(item)

        if item_id not in self._expiration_tasks or not len(self._expiration_tasks[item_id]):
            return

        await ensure_cancelled(self._expiration_tasks[item_id].pop(0))

        if not self._expiration_tasks[item_id]:
            del self._expiration_tasks[item_id]

    async def _remove_all_expiration_handlers(self) -> None:
        await ensure_cancelled_many(itertools.chain.from_iterable(self._expiration_tasks.values()))

        self._expiration_tasks.clear()

    async def _expire_item_shielded(self, item: TItem) -> None:
        # We need to shield this function as by self.remove it will try to cancel itself
        await asyncio.shield(self._expire_item(item))

    async def _expire_item(self, item: TItem) -> None:
        try:
            await self.remove(item)
        except ValueError:
            # Item was removed in the meantime
            return

        if self._on_expired_func:
            await resolve_maybe_awaitable(self._on_expired_func(item))

    async def get(self, *, lock=True) -> TItem:
        async with self._handle_lock(lock):
            item = await super().get(lock=False)

            await self._remove_expiration_task_for_item(item)

            return item

    async def get_all(self, *, lock=True) -> MutableSequence[TItem]:
        async with self._handle_lock(lock):
            items = await super().get_all(lock=False)

            await self._remove_all_expiration_handlers()

            return items

    async def put(self, item: TItem, *, lock=True) -> None:
        async with self._handle_lock(lock):
            await super().put(item, lock=False)

            await self._add_expiration_task_for_item(item)

    async def put_all(self, items: Sequence[TItem], *, lock=True) -> None:
        async with self._handle_lock(lock):
            await super().put_all(items, lock=False)

            await self._remove_all_expiration_handlers()

            await asyncio.gather(
                *[self._add_expiration_task_for_item(item) for item in items],
                return_exceptions=True,
            )

    async def remove(self, item: TItem, *, lock=True) -> None:
        async with self._handle_lock(lock):
            await super().remove(item, lock=False)

            await self._remove_expiration_task_for_item(item)


class BackgroundFillBuffer(ComposableBuffer[TItem]):
    """Composable `Buffer` that adds option to fill buffer in background tasks.

    Background fill will happen only if background tasks are started by calling `.start()`
    and items were requested by `.request()`.
    """

    def __init__(
        self,
        buffer: Buffer[TItem],
        fill_func: Callable[[], Awaitable[TItem]],
        fill_concurrency_size=1,
        on_added_func: Optional[Callable[[TItem], Awaitable[None]]] = None,
    ):
        super().__init__(buffer)

        self._fill_func = fill_func
        self._fill_concurrency_size = fill_concurrency_size
        self._on_added_func = on_added_func

        self._is_started = False
        self._worker_tasks: List[asyncio.Task] = []
        self._workers_semaphore = SingleUseSemaphore()

    async def start(self) -> None:
        if self.is_started():
            raise RuntimeError("Already started!")

        for i in range(self._fill_concurrency_size):
            self._worker_tasks.append(
                create_task_with_logging(
                    self._worker_loop(), trace_id=get_trace_id_name(self, f"worker-{i}")
                )
            )

        self._is_started = True

    async def stop(self) -> None:
        await ensure_cancelled_many(self._worker_tasks)
        self._worker_tasks.clear()
        self._is_started = False

        self._workers_semaphore.reset()

    def is_started(self) -> bool:
        return self._is_started

    async def _worker_loop(self) -> None:
        while True:
            logger.debug("Waiting for fill item request...")

            async with self._workers_semaphore:
                logger.debug("Waiting for fill item request done")

                logger.debug("Adding new item...")

                try:
                    item = await self._fill_func()
                except Exception as e:
                    await self.set_exception(e)
                    logger.debug(
                        "Encountered unexpected exception while adding a new item,"
                        " exception is set and worker loops will be stopped!"
                    )

                    loop = asyncio.get_event_loop()
                    loop.call_soon(lambda: asyncio.create_task(self.stop()))

                    return

                await self.put(item)

            if self._on_added_func:
                await self._on_added_func(item)

            logger.debug("Adding new item done with total of %d items in buffer", self.size())

    async def request(self, count: int) -> None:
        """Request given number of items to be filled in background."""

        await self._workers_semaphore.increase(count)

    def size_with_requested(self) -> int:
        """Return sum of item count stored in buffer and requested to be filled."""

        return self.size() + self._workers_semaphore.get_count_with_pending()

    async def get_requested(self, debounce: timedelta) -> MutableSequence[TItem]:
        """Await for any requested items with given debounce time window, then remove and return \
        all items stored in buffer."""

        await self._buffer.wait_for_any_items()

        try:
            await asyncio.wait_for(
                self._workers_semaphore.finished.wait(), debounce.total_seconds()
            )
        except asyncio.TimeoutError:
            pass

        return await self.get_all()



import asyncio


class SingleUseSemaphore:
    """Class similar to `asyncio.Semaphore` but with more limited count of `.acquire()` calls and\
    exposed counters."""

    def __init__(self, value=0):
        if value < 0:
            raise ValueError("Initial value must be greater or equal to zero!")

        self._value = value

        self._pending = 0
        self._condition = asyncio.Condition()

        self.finished = asyncio.Event()
        if self.locked():
            self.finished.set()

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(self, exc_type, exc, tb):
        self.release()

    def locked(self) -> bool:
        """Return True if there are no more "charges" left in semaphore."""

        return not self._value

    async def acquire(self) -> None:
        """Decrease "charges" counter and increase pending count, or await until there any\
        "charges"."""

        async with self._condition:
            await self._condition.wait_for(lambda: self._value)

            self._value -= 1
            self._pending += 1

    def release(self) -> None:
        """Decrease pending count."""

        if self._pending <= 0:
            raise RuntimeError("Release called too many times!")

        self._pending -= 1

        if self.locked():
            self.finished.set()

    async def increase(self, value: int) -> None:
        """Add given "charges" amount."""

        async with self._condition:
            self._value += value
            self.finished.clear()
            self._condition.notify(value)

    def get_count(self) -> int:
        """Return "charges" count."""

        return self._value

    def get_count_with_pending(self) -> int:
        """Return sum of "charges" and pending count."""

        return self.get_count() + self.get_pending_count()

    def get_pending_count(self) -> int:
        """Return pending count."""

        return self._pending

    def reset(self) -> None:
        """Reset "charges" amount to zero."""

        self._value = 0



import asyncio
import logging
from typing import Dict
from urllib.parse import urlparse

from golem.managers.base import NetworkManager
from golem.node import GolemNode
from golem.resources import DeployArgsType, Network, NewAgreement
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)


class SingleNetworkManager(NetworkManager):
    def __init__(self, golem: GolemNode, ip: str) -> None:
        self._golem = golem
        self._ip = ip
        self._nodes: Dict[str, str] = {}

    @trace_span()
    async def start(self):
        self._network = await Network.create(self._golem, self._ip, None, None)
        await self._network.add_requestor_ip(None)
        self._golem.add_autoclose_resource(self._network)

        await self._golem.event_bus.on(NewAgreement, self._add_provider_to_network)

    @trace_span()
    async def get_node_id(self, provider_id: str) -> str:
        while True:
            node_ip = self._nodes.get(provider_id)
            if node_ip:
                return node_ip

            # TODO: get rid of sleep
            await asyncio.sleep(0.1)

    @trace_span(show_arguments=True, show_results=True)
    async def get_deploy_args(self, provider_id: str) -> DeployArgsType:
        node_ip = await self.get_node_id(provider_id)
        return self._network.deploy_args(node_ip)

    @trace_span(show_arguments=True, show_results=True)
    async def get_provider_uri(self, provider_id: str, protocol: str = "http") -> str:
        node_ip = await self.get_node_id(provider_id)
        url = self._network.node._api_config.net_url
        net_api_ws = urlparse(url)._replace(scheme=protocol).geturl()
        connection_uri = f"{net_api_ws}/net/{self._network.id}/tcp/{node_ip}/22"
        return connection_uri

    @trace_span(show_arguments=True)
    async def _add_provider_to_network(self, event: NewAgreement):
        await event.resource.get_data()
        provider_id = event.resource.data.offer.provider_id
        assert provider_id is not None  # TODO handle this case better
        logger.info(f"Adding provider {provider_id} to network")
        self._nodes[provider_id] = await self._network.create_node(provider_id)



from golem.managers.base import WorkContext


async def default_on_activity_start(context: WorkContext):
    batch = await context.create_batch()
    batch.deploy()
    batch.start()
    await batch()

    # TODO: After this function call we should have check if activity is actually started


async def default_on_activity_stop(context: WorkContext):
    await context.terminate()

    # TODO: After this function call we should have check if activity is actually terminated



from golem.resources.agreement import Agreement
from golem.resources.proposal.proposal import Proposal


async def default_negotiate(proposal: Proposal) -> Proposal:
    """Negotiate a new :any:`Proposal` based on a given :any:`Proposal`."""
    our_response = await proposal.respond()
    async for their_response in our_response.responses():
        return their_response
    raise Exception(f"Negotiations based on {proposal} failed")


async def default_create_agreement(proposal: Proposal) -> Agreement:
    """Create a both-sides approved :any:`Agreement` based on a given :any:`Proposal`."""
    agreement = await proposal.create_agreement()
    await agreement.confirm()

    approved = await agreement.wait_for_approval()
    if approved:
        return agreement
    raise Exception(f"Agreement {agreement} created from {proposal} was not approved")
